import click
import tables
import pandas as pd
from rich import print


@click.command()
@click.argument('hdf-file', type=click.Path(exists=True),
                metavar='[HDF-DATA]')
@click.option('--rows', '-r', type=int, default=50)
@click.option('--start', '-s', type=int, default=0)
def show_hdf(hdf_file, rows, start):
    pd.set_option('display.max_rows', None)
    pd.set_option('display.max_columns', None)
    pd.set_option('display.width', None)
    pd.set_option('display.max_colwidth', None)
    tb = tables.open_file(hdf_file)
    data = tb.root.data.measurements
    df = pd.DataFrame.from_records(data.read())
    print(df.iloc[start:start+rows])



from pathlib import Path


def flatten(nested_list: list) -> list:
    flattened_list = []
    for elem in nested_list:
        if isinstance(elem, list):
            flattened_list += flatten(elem)
        else:
            flattened_list.append(elem)
    return flattened_list


def test_flatten():
    tlist = [1, 2, 3, [4, 5, 6], [[7, 8], [9, 10], 11, 12], [13], 14, 15]
    flattened_list = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15]
    assert flatten(tlist) == flattened_list

    tlist = [1, 2, 3, []]
    flattened_list = [1, 2, 3]
    assert flatten(tlist) == flattened_list


def get_procedure_with_name(procedure_names,
                            procedures, workflows):
    avail_proc_names = [p['name'] for p in procedures]
    workflow_names = [w['name'] for w in workflows]
    # check if the name given is in the procedures list.
    procedures = []
    for procedure_name in procedure_names:
        if procedure_name in avail_proc_names:
            procedure_index = procedure_names.index(procedure_name)
            procedure = procedures[procedure_index]
            procedures.append(procedure)
        # check if the name is in the list of workflows
        elif procedure_name in workflow_names:
            workflow_index = workflow_names.index(procedure_name)
            workflow = workflows[workflow_index]
            procedures.append([get_procedure_with_name(task_name,
                                                       procedures,
                                                       workflows)
                               for task_name in workflow['tasks']])
    return flatten(procedures)


def test_get_procedure_with_name():
    # test the base case where the procedure name is not nested
    root_cfg_file = Path('../../tests/configuration/main_config.yaml')
    procedure_name = 'timewalk_scan'
    procedures = get_procedure_with_name(
            root_cfg_file,
            procedure_name)
    assert procedures['name'] == procedure_name
    assert isinstance(procedures['parameters'], list)
    assert len(procedures['parameters']) == 1
    assert len(procedures['parameters'][0]) == 2048
    assert isinstance(procedures['parameters'][0][0], dict)
    assert isinstance(procedures['target_settings']['power_on_default'], dict)

    procedure_name = 'master_tdc_daq'
    procedures = get_procedure_with_name(
            root_cfg_file,
            procedure_name)
    assert isinstance(procedures, list)
    assert len(procedures) == 2
    assert procedures[0]['name'] == 'master_tdc_SIG_calibration_daq'
    assert procedures[1]['name'] == 'master_tdc_REF_calibration_daq'
    assert 'parameters' in procedures[0]
    assert 'parameters' in procedures[1]
    assert 'data_columns' in procedures[0]
    assert len(procedures[0]['data_columns']) == 13



from luigi.freezing import FrozenOrderedDict


def unfreeze(config):
    """
    unfreeze the dictionary that is frozen to serialize it
    and send it from one luigi task to another
    """
    try:
        unfrozen_dict = {}
        for key in config.keys():
            unfrozen_dict[key] = unfreeze(config[key])
        return unfrozen_dict
    except AttributeError:
        try:
            unfrozen_list = []
            if not isinstance(config, str):
                for elem in config:
                    unfrozen_list.append(unfreeze(elem))
                return unfrozen_list
            return config
        except TypeError:
            return config


def test_unfreeze():
    input_dict = FrozenOrderedDict({'a': 1, 'b': (1, 2, 3),
                                    'd': FrozenOrderedDict({'e': 4})})
    expected_dict = {'a': 1, 'b': [1, 2, 3],
                     'd': {'e': 4}}
    assert expected_dict == unfreeze(input_dict)



import os
import multiprocessing as mp
import math
import logging
import queue
import threading
from pathlib import Path
import click
from rich.progress import Progress
from rich.progress import SpinnerColumn
from rich.progress import MofNCompleteColumn
from rich.progress import TextColumn
from rich.progress import BarColumn
from rich.progress import TimeElapsedColumn, TimeRemainingColumn
import yaml
from . import config_utilities as cfu
from . import dict_utils as dctu
from .gen_configurations import pipelined_generate_patches
from .gen_configurations import pipelined_generate_full_run_config
from .gen_configurations import generate_tool_configurations
from .acquire_data import pipelined_acquire_data
from .postprocessing_queue import unpack_data, frack_data


_log_level_dict = {
    "DEBUG": logging.DEBUG,
    "INFO": logging.INFO,
    "WARNING": logging.WARNING,
    "ERROR": logging.ERROR,
    "CRITICAL": logging.CRITICAL,
}


@click.command()
@click.argument("config", type=click.Path(exists=True), metavar="[ROOT CONFIG FILE]")
@click.argument(
    "netcfg", type=click.Path(exists=True), metavar="[NERTWORK CONFIG FILE]"
)
@click.argument("procedure", type=str, metavar="[PROCEDURE NAME]")
@click.argument(
    "output_dir", type=click.Path(dir_okay=True), metavar="[OUTPUT DIRECTORY]"
)
@click.option(
    "--log",
    "-l",
    type=str,
    default="daq.log",
    help="Enable logging and append logs to the filename passed to " "this option",
)
@click.option(
    "--loglevel",
    "-v",
    default="INFO",
    type=click.Choice(
        ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"], case_sensitive=False
    ),
)
@click.option(
    "--root/--no-root",
    default=False,
    help="keep the rootfile generated as intermediary",
)
@click.option(
    "--full_conf_generators",
    "-f",
    type=int,
    default=2,
    help="set how many full-cfg generators to run, default 1",
)
@click.option(
    "--unpack_tasks",
    "-u",
    default=2,
    type=int,
    help="number of unpackers/frackers to run in parallel",
)
@click.option(
    "--frack_tasks",
    "-f",
    default=7,
    type=int,
    help="number of unpackers/frackers to run in parallel",
)
@click.option(
    "--compression",
    "-c",
    default=2,
    type=int,
    help="Set the compression for the hdf file, 0 = no compression"
    " 9 = best compression",
)
@click.option(
    "--keep/--no-keep",
    default=False,
    help="Keep the already aquired data, defaults to False",
)
@click.option(
    "--readback/--no-readback",
    default=False,
    help="Tell the sc-server to read back the register value" "written to the roc",
)
@click.option(
    "--full-readback/-no-full-readback",
    default=False,
    help="Enable a Full read of the ROC parameters after every run",
)
def main(
    config,
    netcfg,
    procedure,
    output_dir,
    log,
    loglevel,
    root,
    full_conf_generators,
    unpack_tasks,
    frack_tasks,
    compression,
    keep,
    readback,
    full_readback,
):
    config = click.format_filename(config)
    # generate the conifgurations
    if log:
        logging.basicConfig(
            filename="full-daq.log",
            level=loglevel,
            format="[%(asctime)s] %(levelname)-10s:" "%(name)-50s %(message)s",
        )
    config = click.format_filename(config)
    try:
        procedure, (
            system_default_config,
            system_init_config,
            run_configs,
            run_count,
        ) = cfu.get_procedure_configs(
            main_config_file=config,
            procedure_name=procedure,
            calibration=None,
            diff=True,
        )
    except ValueError as err:
        print(f"The procedure with name: {err.args[1]} could not be found,")
        print("Available procedures are:")
        for pname in err.args[2]:
            print(f"\t {pname}")
        exit(1)

    # create the output directory and the initial files
    output_dir = Path(output_dir)
    if not os.path.exists(output_dir):
        os.mkdir(output_dir)

    generate_tool_configurations(
        output_dir, netcfg, procedure, system_default_config, system_init_config
    )
    with open(netcfg, "r", encoding="utf-8") as nw_file:
        network_config = yaml.safe_load(nw_file.read())

    # get the data needed for the postprocessing
    raw_data = procedure["mode"] == "full"
    data_columns = procedure["data_columns"]
    try:
        data_format = procedure["data_format"]
    except KeyError:
        data_format = "raw"
    characterisation_mode = data_format == "characterisation"

    # generate the configurations for the runs
    num_digits = math.ceil(math.log(run_count, 10))
    full_config = dctu.update_dict(system_default_config, system_init_config)

    # event telling all the subprocesses to stop
    stop = mp.Event()

    # queue that holds the filenames for a run indicating the
    # run config has been generated
    patch_gen_out_full_gen_in = queue.Queue()
    patch_gen_progress_queue = queue.Queue()
    patch_gen_done = threading.Event()

    # queues and signals relating to the full config generation
    fcm = mp.Manager()
    full_gen_out_daq_in = fcm.Queue()
    full_config_gen_progress = fcm.Queue()
    full_config_gen_done = threading.Event()

    # daq
    daq_out_unpack_in = queue.Queue()
    daq_progress = queue.Queue()
    daq_done = threading.Event()
    daq_active = False
    daq_system_initialized = threading.Event()

    # unpacking
    unpack_out_frack_in = queue.Queue()
    unpack_progress = queue.Queue()
    unpack_done = threading.Event()

    # fracking
    frack_progress = queue.Queue()
    fracking_done = threading.Event()

    config_gen_thread = threading.Thread(
        target=pipelined_generate_patches,
        args=(
            output_dir,  # directory to place the
            run_configs,
            patch_gen_out_full_gen_in,
            patch_gen_progress_queue,
            patch_gen_done,
            num_digits,
            stop,
        ),
    )
    full_config_gen_thread = threading.Thread(
        target=pipelined_generate_full_run_config,
        args=(
            patch_gen_done,
            patch_gen_out_full_gen_in,
            full_config,
            full_gen_out_daq_in,
            full_config_gen_progress,
            full_config_gen_done,
            full_conf_generators,
            stop,
        ),
    )
    # acquire the data in a separate thread
    daq_thread = threading.Thread(
        target=pipelined_acquire_data,
        args=(
            full_gen_out_daq_in,
            daq_progress,
            daq_out_unpack_in,
            full_config_gen_done,
            daq_system_initialized,
            daq_done,
            stop,
            network_config,
            system_init_config,
            output_dir,
            keep,
            readback,
            procedure["run_start_tdc_procedure"],
            full_readback,
        ),
    )

    # start the unpack_tasks that turn the raw data into the root file
    unpack_thread = threading.Thread(
        target=unpack_data,
        args=(
            daq_out_unpack_in,
            unpack_progress,
            unpack_out_frack_in,
            daq_done,
            unpack_done,
            max(1, unpack_tasks),
            raw_data,
            characterisation_mode,
        ),
    )

    # frack the data (last step)
    frack_thread = threading.Thread(
        target=frack_data,
        args=(
            unpack_out_frack_in,
            frack_progress,
            unpack_done,
            fracking_done,
            max(1, frack_tasks),
            raw_data,
            root,
            compression,
            data_columns,
        ),
    )
    config_gen_thread.start()
    full_config_gen_thread.start()
    daq_thread.start()
    unpack_thread.start()
    frack_thread.start()
    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        BarColumn(bar_width=None),
        MofNCompleteColumn(),
        TimeElapsedColumn(),
        TimeRemainingColumn(),
    ) as progress:
        run_config_progress_bar = progress.add_task(
            "Generating run configurations", total=run_count
        )
        full_config_progress_bar = progress.add_task(
            "Generating full config for fracker", total=run_count
        )
        daq_progbar = progress.add_task(
            "Acquiring Data from the Test System", total=run_count, start=False
        )
        unpack_progress_bar = progress.add_task("Unpacking data", total=run_count)
        frack_progress_bar = progress.add_task("fracking data", total=run_count)
        while not fracking_done.is_set():
            try:
                _ = patch_gen_progress_queue.get(block=False)
                progress.update(run_config_progress_bar, advance=1)
            except queue.Empty:
                pass
            try:
                _ = full_config_gen_progress.get(block=False)
                progress.update(full_config_progress_bar, advance=1)
            except queue.Empty:
                pass
            if daq_system_initialized.is_set() and not daq_active:
                progress.start_task(daq_progbar)
                daq_active = True
            try:
                if daq_active:
                    _ = daq_progress.get(block=False)
                    progress.update(daq_progbar, advance=1)
            except queue.Empty:
                pass
            try:
                _ = unpack_progress.get(block=False)
                progress.update(unpack_progress_bar, advance=1)
            except queue.Empty:
                pass
            try:
                _ = frack_progress.get(block=False)
                progress.update(frack_progress_bar, advance=1)
            except queue.Empty:
                pass
        progress.refresh()
    config_gen_thread.join()
    full_config_gen_thread.join()
    daq_thread.join()
    unpack_thread.join()
    frack_thread.join()



import threading
import queue
import logging
import os
import sys
from pathlib import Path
from time import sleep
import glob
from typing import List
import yaml
from rich.progress import Progress
from rich.progress import SpinnerColumn
from rich.progress import MofNCompleteColumn
from rich.progress import TextColumn
from rich.progress import BarColumn
from rich.progress import TimeElapsedColumn, TimeRemainingColumn
import click
from . import analysis_utilities as anu


def unpack_data(
    raw_data_queue: queue.Queue,
    unpack_reporting_queue: queue.Queue,
    root_data_queue: queue.Queue,
    data_taking_done: threading.Event,
    unpacking_done: threading.Event,
    max_parallel_unpackers: int,
    raw_data: bool,
    characterisation_mode: bool,
):
    logger = logging.getLogger("unpack-data-thread")
    running_tasks: List = []
    while (
        not data_taking_done.is_set()
        or not raw_data_queue.empty()
        or len(running_tasks) != 0
    ):
        if len(running_tasks) < max_parallel_unpackers:
            try:
                (
                    patch_path,
                    full_config_path,
                    raw_path,
                    unpack_path,
                    fracked_path,
                    readback_config_path,
                ) = raw_data_queue.get(block=False)
                logger.info("Received data to unpack," f"{os.path.basename(raw_path)}")
                running_tasks.append(
                    (
                        anu.start_unpack(
                            raw_path,
                            unpack_path,
                            logging.getLogger(f"unpack-{os.path.basename(raw_path)}"),
                            raw_data=raw_data,
                            characMode=characterisation_mode,
                        ),
                        raw_path,
                        unpack_path,
                        fracked_path,
                        full_config_path,
                    )
                )
            except queue.Empty:
                pass
        del_indices = []
        for i, (
            task,
            raw_path,
            unpack_path,
            fracked_path,
            full_config_path,
        ) in enumerate(running_tasks):
            returncode = task.poll()
            if returncode is None:
                continue
            logger.debug(f"currently running tasks {running_tasks}")
            unpack_reporting_queue.put(1)
            del_indices.append(i)
            if returncode == 0:
                logger.info("created root file from " f"{os.path.basename(raw_path)}")
                logger.info("putting fracker task on the queue")
                root_data_queue.put(
                    (raw_path, unpack_path, fracked_path, full_config_path)
                )
            if returncode > 0:
                logger.error(
                    f"failed to created root file from {raw_path}, "
                    f"unpacker returned error code {returncode}"
                )
                os.remove(raw_path)
            if returncode < 0:
                logger.error(
                    f"failed to created root file from {raw_path}, "
                    "unpacker got terminated"
                )
                os.remove(raw_path)
            raw_data_queue.task_done()
        running_tasks = list(
            filter(lambda x: running_tasks.index(x) not in del_indices, running_tasks)
        )
        sleep(0.05)
    unpacking_done.set()


def frack_data(
    frack_data_queue: queue.Queue,
    frack_report_queue: queue.Queue,
    previous_task_complete: threading.Event,
    fracking_done: threading.Event,
    max_parallel_frackers: int,
    raw_data: bool,
    keep_root: bool,
    compression: int,
    columns: list,
):
    running_tasks = []
    logger = logging.getLogger("data-fracking-thread")
    while (
        not previous_task_complete.is_set()
        or not frack_data_queue.empty()
        or not len(running_tasks) == 0
    ):
        if len(running_tasks) < max_parallel_frackers:
            try:
                (
                    raw_path,
                    unpack_path,
                    fracked_path,
                    full_config_path,
                ) = frack_data_queue.get(block=False)
                logger.info(
                    "Received data to frack, " f"{os.path.basename(unpack_path)}"
                )
                running_tasks.append(
                    (
                        anu.start_compiled_fracker(
                            unpack_path,
                            fracked_path,
                            full_config_path,
                            raw_data,
                            columns,
                            compression,
                            logger,
                        ),
                        raw_path,
                        unpack_path,
                        fracked_path,
                        full_config_path,
                    )
                )
            except queue.Empty:
                pass
        del_indices = []
        for i, (
            task,
            raw_path,
            unpack_path,
            fracked_path,
            full_config_path,
        ) in enumerate(running_tasks):
            returncode = task.poll()
            if returncode is None:
                continue
            logger.debug(f"currently running tasks {running_tasks}")
            frack_report_queue.put(1)
            del_indices.append(i)
            if returncode == 0:
                logger.info(f"created {os.path.basename(fracked_path)}")

            if returncode > 0:
                logger.error(
                    "failed to created hdf file from "
                    f"{os.path.basename(unpack_path)}, "
                    f"fracker returned error code {returncode}"
                )
                os.remove(raw_path)
            if returncode < 0:
                logger.error(
                    "failed to created hdf file from "
                    f"{os.path.basename(raw_path)}, "
                    "unpacker got terminated"
                )
                os.remove(raw_path)
            if not keep_root:
                os.remove(unpack_path)
            frack_data_queue.task_done()
        running_tasks = list(
            filter(lambda x: running_tasks.index(x) not in del_indices, running_tasks)
        )
        sleep(0.05)
    fracking_done.set()


_log_level_dict = {
    "DEBUG": logging.DEBUG,
    "INFO": logging.INFO,
    "WARNING": logging.WARNING,
    "ERROR": logging.ERROR,
    "CRITICAL": logging.CRITICAL,
}


@click.command
@click.argument(
    "output_dir",
    type=click.Path(dir_okay=True),
    metavar="[Location containing the config and data]",
)
@click.option(
    "--log",
    type=str,
    default="post.log",
    help="Enable logging and append logs to the filename passed to " "this option",
)
@click.option(
    "--loglevel",
    default="INFO",
    type=click.Choice(
        ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"], case_sensitive=False
    ),
)
@click.option(
    "--root/--no-root",
    default=False,
    help="keep the rootfile generated as intermediary",
)
@click.option(
    "--unpack_tasks",
    default=1,
    type=int,
    help="number of unpacking tastks to run in parallel",
)
@click.option(
    "--frack_tasks", default=3, type=int, help="number of frackers to run in parallel"
)
@click.option(
    "--compression",
    "-c",
    default=3,
    type=int,
    help="Set the compression for the hdf file, 0 = no compression"
    " 9 = best compression",
)
def main(output_dir, log, loglevel, root, unpack_tasks, frack_tasks, compression):
    if log is not None:
        logging.basicConfig(
            filename=log,
            level=_log_level_dict[loglevel],
            format="[%(asctime)s] %(levelname)s:" "%(name)-50s %(message)s",
        )
    # read in the information for the postprocessing from the config files
    output_dir = Path(output_dir)
    with open(output_dir / "postprocessing_config.yaml") as pcf:
        procedure = yaml.safe_load(pcf.read())
        try:
            data_columns = procedure["data_columns"]
        except KeyError:
            print("The procedure needs to specify data columns")
            sys.exit(1)
        try:
            mode = procedure["mode"]
        except KeyError:
            mode = "summary"
        raw_data = True if mode == "full" else False
        try:
            data_format = procedure["data_format"]
        except KeyError:
            data_format = "raw"
        characterisation_mode = True if data_format == "characterisation" else False

    # prepare the threading environment
    raw_data_queue = queue.Queue()
    root_data_queue = queue.Queue()
    unpack_reporting_queue = queue.Queue()
    fracker_reporting_queue = queue.Queue()
    data_taking_done = threading.Event()
    unpack_done = threading.Event()
    fracking_done = threading.Event()

    # initilize the threads
    unpack_thread = threading.Thread(
        target=unpack_data,
        args=(
            raw_data_queue,
            unpack_reporting_queue,
            root_data_queue,
            data_taking_done,
            unpack_done,
            max(1, unpack_tasks),
            raw_data,
            characterisation_mode,
        ),
    )
    frack_thread = threading.Thread(
        target=frack_data,
        args=(
            root_data_queue,
            fracker_reporting_queue,
            unpack_done,
            fracking_done,
            max(1, frack_tasks),
            raw_data,
            root,
            compression,
            data_columns,
        ),
    )

    unpack_thread.start()
    frack_thread.start()

    # read in the data and start the postprocessing
    run_config_paths = glob.glob(str(output_dir.absolute()) + "/run_*_config.yaml")
    run_raw_data_paths = glob.glob(str(output_dir.absolute()) + "/run_*_data.raw")
    run_config_paths = sorted(run_config_paths, key=lambda x: int(x.split("_")[-2]))
    run_raw_data_paths = sorted(run_raw_data_paths, key=lambda x: int(x.split("_")[-2]))
    run_root_data_paths = list(
        map(lambda x: os.path.splitext(x)[0] + ".root", run_raw_data_paths)
    )
    run_hdf_data_paths = list(
        map(lambda x: os.path.splitext(x)[0] + ".h5", run_raw_data_paths)
    )
    full_config_paths = list(
        map(lambda x: Path(os.path.splitext(x)[0] + "_full.yaml"), run_config_paths)
    )
    run_count = len(run_config_paths)
    for full_config_path, raw_data_path, root_data_path, hdf_data_path in zip(
        full_config_paths, run_raw_data_paths, run_root_data_paths, run_hdf_data_paths
    ):
        if not full_config_path.exists():
            print(f"Full config file {full_config_path} not found")
            data_taking_done.set()
            unpack_done.set()
            unpack_thread.join()
            frack_thread.join()
            sys.exit(1)
        raw_data_queue.put(
            (
                Path(raw_data_path),
                Path(root_data_path),
                Path(hdf_data_path),
                Path(full_config_path),
            )
        )
    data_taking_done.set()
    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        BarColumn(bar_width=None),
        MofNCompleteColumn(),
        TimeElapsedColumn(),
        TimeRemainingColumn(),
    ) as progress:
        unpack_progress_bar = progress.add_task("Unpacking data", total=run_count)
        frack_progress_bar = progress.add_task("fracking data", total=run_count)
        while not fracking_done.is_set():
            try:
                _ = unpack_reporting_queue.get(block=False)
                progress.update(unpack_progress_bar, advance=1)
            except queue.Empty:
                pass
            try:
                _ = fracker_reporting_queue.get(block=False)
                progress.update(frack_progress_bar, advance=1)
            except queue.Empty:
                pass
            sleep(0.05)
        unpack_thread.join()
        frack_thread.join()



from schema import Schema, And, Use, Or, Optional
from pathlib import Path
import yaml
import os


current_path = Path('.')


def set_current_path(path):
    global current_path
    current_path = path


def load_configuration(config_path):
    """
    load the configuration dictionary from a yaml file

    :raises: ConfigFormatError if the input cannot be parsed by the yaml
    parser
    """
    with open(config_path, 'r', encoding='utf-8') as config_file:
        return yaml.safe_load(config_file.read())


def full_path(path):
    global current_path
    fpath = Path(current_path) / path
    if fpath.exists():
        return fpath.resolve()
    raise FileNotFoundError(f'{fpath.resolve()} cannot be found')


def generate_values(rdict):
    stop = rdict['stop']
    try:
        start = rdict['start']
    except KeyError:
        start = 0
    try:
        step = rdict['step']
    except KeyError:
        step = 1
    return list(range(start, stop, step))


def resolve_library(path):
    global current_path
    path = Path(path)
    path = Path(current_path) / path
    if not os.path.exists(path.absolute()):
        raise FileNotFoundError(
                f"The config file {path.resolve()} cant be found")
    tmp_path = current_path
    current_path = os.path.dirname(path)
    validated_lib = library_file.validate(load_configuration(path))
    current_path = tmp_path
    return validated_lib


systems_settings = Schema(
        {'default': Or(
            [Use(full_path,
                 error="a default config file could not be found")],
            And(str, Use(full_path,
                         error="a default config file could not be found"))),
         Optional('init', default=[]): Or(
             [Use(full_path,
                  error="an init config file coule not be found")],
             And(str, Use(full_path,
                          error="an init config file coule not be found"))),
         Optional('override', default={}): dict})


parameter_range = Schema(
        Or(
            And(dict, Use(generate_values,
                          error='values field is not a range description')),
            [dict, int], error="values is neither a list values nor a "
                               "range description")
         )

parameter = Schema(
        Or({'key': [list, str, int],
            'values': parameter_range},
           {'template': str,
            'values': parameter_range,
            'default': str},
           error="A parameter must have a 'template', "
                 "'values' and 'default' field "
                 " or 'key' and 'values' field")
         )

daq_config = Schema(
        {'name': str,
         'type': 'daq',
         'system_settings': systems_settings,
         Optional('calibration', default=None): str,
         Optional('merge', default=True): bool,
         Optional('run_start_tdc_procedure', default=False): bool,
         Optional('repeat', default=1): int,
         Optional('mode', default='summary'): Or('summary', 'full'),
         Optional('data_format', default='raw'): Or('raw', 'characterisation'),
         Optional('parameters', default=None): [parameter],
         Optional('data_columns', default=[]): [str]
         }
)

analysis_config = Schema(
        {'name': str,
         'type': 'analysis',
         'daq': str,
         Optional('compatible_modes', default='summary'):
            Or('summary', 'full',
               error="compatible modes are either 'summary' or 'full'"),
         Optional('provides_calibration', default=False): bool,
         Optional('module_name'): str,
         Optional('parameters', default={}): dict
         }
)

procedure_config = Schema(Or(analysis_config, daq_config))

workflow_config = Schema({'name': str,
                          'tasks': [str]})

library_file = Schema([procedure_config])


main_config = Schema({
    Optional('libraries', default=[]): [Use(resolve_library)],
    Optional('procedures', default=[]): [procedure_config],
    Optional('workflows', default=[]): [workflow_config]
    })



import luigi
import importlib
import os
import sys
import yaml
from copy import deepcopy
from pathlib import Path
from .analysis_utilities import read_dataframe_chunked
from .analysis_utilities import read_whole_dataframe
from .config_utilities import unfreeze, update_dict
from .errors import OutputError


class Distillery(luigi.Task):
    """ Task that encapsulates analysis tasks and makes them executable
    inside the Datenraffinerie
    """
    system_state = luigi.DictParameter(significant=True)
    priority = luigi.OptionalParameter(significant=False, default=0)

    def requires(self):
        from .valve_yard import ValveYard
        """ Determin which analysis needs to be run to produce
        the data for the analysis
        :returns: The acquisition procedure needed to produce the data
        """
        subtask_state = deepcopy(self.system_state)
        subtask_state['output_path'] = os.path.dirname(
                self.system_state['output_path'])
        subtask_state['name'] = self.system_state['procedure']['daq']
        del subtask_state['procedure']
        return ValveYard(subtask_state)

    def output(self):
        """ Define the files that are produced by the analysis
        :returns: dictionary with the keys 'summary', 'calibration'
            and 'plots', where 'summary's value is a string for a
            relative path to the summary, same as calibration is a
            relative path to the calibration yaml file
            and the value associated to 'plots' is a list of relative
            paths. All paths should be strings.
        """
        analysis_module_path = self.system_state['analysis_module_path']
        output_path = self.system_state['output_path']
        python_module = self.system_state['procedure']['python_module_name']
        parameters = self.system_state['procedure']['parameters']
        analysis = self.import_analysis(analysis_module_path,
                                        python_module)
        analysis_parameters = unfreeze(parameters)
        analysis = analysis(analysis_parameters)
        output = {}
        output['full-calibration'] = self.input()['full-calibration']
        for key, paths in analysis.output().items():
            if paths is None:
                continue
            try:
                if len(paths) == 0:
                    continue
            except TypeError:
                pass
            if key == 'plots':
                try:
                    if len(paths) == 0:
                        continue
                    output['plots'] = [luigi.LocalTarget((Path(output_path)
                                                          / path).resolve())
                                       for path in paths]
                except TypeError as err:
                    raise OutputError('The plots output must be a list ' +
                                      'of paths. Use an empty list if ' +
                                      'no plots are generated') from err
            elif key == 'calibration':
                output['full-calibration']\
                        = (Path(output_path) / paths).resolve()
            else:
                if isinstance(paths, list):
                    output[key] = [Path(output_path) / path for path in paths]
                else:
                    output[key] = Path(output_path) / paths
        return output

    def run(self):
        """ perform the analysis using the imported distillery
        :returns: TODO

        """
        analysis_module_path = self.system_state['analysis_module_path']
        output_path = self.system_state['output_path']
        python_module = self.system_state['procedure']['python_module_name']
        parameters = self.system_state['procedure']['parameters']
        analysis = self.import_analysis(analysis_module_path,
                                        python_module)
        analysis_parameters = unfreeze(parameters)

        # import the class definition
        analysis = self.import_analysis(analysis_module_path,
                                        python_module)
        # instantiate an analysis object from the imported analysis class
        analysis = analysis(analysis_parameters)

        event_mode = self.system_state['event_mode']
        # open and read the data
        if not event_mode:
            analysis.run(read_whole_dataframe(self.input()['data'].path),
                         output_path)
        else:
            in_files = []
            for dfile in self.input()['data']:
                if isinstance(dfile, list):
                    in_files += dfile
                else:
                    in_files.append(dfile)
            data_iter = read_dataframe_chunked([i.path for i in in_files])
            analysis.run(data_iter, output_path)

        # merge the calibrations
        if self.output()['full-calibration'].path\
                != self.input()['full-calibration'].path:
            with self.input()['full-calibration'].open('r') as prev_calib_file:
                previous_calib = yaml.safe_load(prev_calib_file.read())
            with self.output()['full-calibration'].open('w') as new_calib_file:
                new_calib = yaml.safe_load(new_calib_file.read())
                update_dict(previous_calib, new_calib, in_place=True)
                new_calib_file.truncate(0)
                new_calib_file.write(yaml.safe_dump(previous_calib))

    @staticmethod
    def import_analysis(distillery_path: str, name: str):
        """ Import the distillery for the analysis.

        :distillery_path: The path in which to find the distilleries
            module
        :name: The name of the distillery to load
        :returns: the distillery loaded into the local namespace
        """
        if distillery_path is not None:
            pathstr = str(Path(distillery_path).resolve())
            pythonpath_entry = os.path.split(pathstr)[0]
            module_name = os.path.split(pathstr)[1]
            sys.path.append(pythonpath_entry)
            i = importlib.import_module(module_name)
            distillery = getattr(i, name)
        else:
            import datenraffinerie_distilleries as distilleries
            distillery = getattr(distilleries, name)
        return distillery



import zmq
import sys
import bson
import yaml
import logging
import click
import shutil
import subprocess as sp
from pathlib import Path
from .control_adapter import DAQSystem
from hgcroc_configuration_client.client import Client as SCClient
from schema import Schema, Or
from schema import SchemaError
from .errors import DAQError
import uuid
from time import sleep
from time import ctime
import os


class DAQCoordCommand():
    valid_commands = ['acquire lock', 'shutdown', 'release lock',
                      'reset_target', 'read_target', 'initialize',
                      'measure', 'configure']
    schema = Schema(Or({
        'command': Or('acquire lock', 'shutdown'),
        'locking_token': None,
        'config': None
    }, {
        'command': 'release lock',
        'locking_token': str,
        'config': None
    }, {
        'command': 'reset_target',
        'locking_token': str,
        'config': None,
    }, {
        'command': 'read_target',
        'locking_token': str,
        'config': Or(str, None)
    }, {
        'command': 'initialize',
        'locking_token': str,
        'config': str
    }, {
        'command': 'measure',
        'locking_token': str,
        'config': str,
        'readback': bool
    }, {
        'command': 'configure',
        'locking_token': str,
        'config': str,
        'readback': bool
    }))

    def __init__(self, command: str, locking_token=None,
                 config=None, readback: bool = False):
        self.command = command if command in self.valid_commands else None
        if command in self.valid_commands[2:] and locking_token is None:
            raise DAQError(f"For a message with the command: {command}, a "
                           "locking token is required")
        self.locking_token = locking_token
        if command in self.valid_commands[5:] and config is None:
            raise DAQError("For a message of type 'initialize', 'configure"
                           "or 'measure' a configuration dict is required")
        if not isinstance(readback, bool):
            raise DAQError('The measurement Command needs a readback '
                           'argument that is a bool')
        self.readback = readback
        self.config = config

    def serialize(self):
        config = None
        if self.config is not None:
            config = yaml.dump(self.config)
        msg_dict = {'command': self.command,
                    'locking_token': self.locking_token,
                    'config': config}
        if self.command in self.valid_commands[6:]:
            msg_dict['readback'] = self.readback
        return bson.encode(msg_dict)

    @staticmethod
    def parse(message: bytes):
        try:
            message_dict = bson.decode(message)
        except:
            raise DAQError('Unable to decode bson message')
        try:
            valid_message = DAQCoordCommand.schema.validate(message_dict)
        except SchemaError as err:
            raise ValueError(
                'Message validation failed for message: '
                f'{message_dict}\n Got Error: {err}')
        parsed_config = None
        if message_dict['config'] is not None:
            parsed_config = yaml.safe_load(valid_message['config'])
        try:
            readback = valid_message['readback']
        except KeyError:
            readback = False
        return DAQCoordCommand(command=valid_message['command'],
                               locking_token=valid_message['locking_token'],
                               config=parsed_config,
                               readback=readback)


class DAQCoordResponse():
    valid_responses = ['error', 'data', 'lock', 'ack', 'access denied',
                       'target_config']
    schema = Schema(Or({
        'type': 'error',
        'content': str
    }, {
        'type': 'data',
        'content': bytes
    }, {
        'type': 'lock',
        'content': Or(str, None)
    }, {
        'type': Or('ack', 'access denied'),
        'content': None
    }, {
        'type': 'target_config',
        'content': str
    })
    )

    def __init__(self, type: str, content=None):
        if type not in self.valid_responses:
            raise DAQError('daq response not valid')
        self.type = type
        if type == 'target_config' and not isinstance(content, dict):
            raise DAQError('The target config needs to be a dict')
        self.content = content
        if type == 'data' and not isinstance(content, bytes):
            raise DAQError('The data needs to be of type bytes')

    def serialize(self, logger):
        if self.type == 'error':
            if self.content is None:
                raise DAQError('daq error response was attempted but'
                               'no error was given')
            else:
                return bson.encode(
                    {'type': 'error', 'content': self.content})
        if self.type == 'data':
            if self.content is None:
                raise DAQError('No data provided to the daq response')
            else:
                logger.debug(f'binary Data is {len(self.content)} bytes long')
                return bson.encode(
                    {'type': 'data',
                     'content': bson.Binary(self.content, subtype=0)})
        if self.type == 'lock':
            return bson.encode(
                {'type': 'lock', 'content': str(self.content)})
        if self.type == 'ack':
            return bson.encode({'type': 'ack', 'content': None})
        if self.type == 'target_config':
            return bson.encode({'type': 'target_config',
                                'content': yaml.safe_dump(self.content)})

    @staticmethod
    def parse(message: bytes, logger):
        rsp_dict = bson.decode(message)
        if rsp_dict['type'] == 'data':
            logger.debug(f'Received {len(rsp_dict["content"])} bytes of data')
        valid_message = DAQCoordResponse.schema.validate(rsp_dict)
        if valid_message['type'] == 'target_config':
            valid_message['content'] = yaml.safe_load(valid_message['content'])
        return DAQCoordResponse(type=valid_message['type'],
                                content=valid_message['content'])

    def __str__(self):
        return f'{self.type}: {self.content}'


class DAQCoordClient():
    def __init__(self, network_config: dict):
        self.logger = logging.getLogger('daq_coord_client')
        self.lock = None
        self.zmq_transaction_running = False
        self.raw_response = None
        self.io_context = zmq.Context()
        self.socket = self.io_context.socket(zmq.REQ)
        socket_address = \
            f"tcp://{network_config['daq_coordinator']['hostname']}:" +\
            f"{network_config['daq_coordinator']['port']}"
        self.socket.connect(socket_address)
        self.logger.info('Initializing')
        response = DAQCoordResponse(type='lock', content=None)
        try_count = 20
        while response.content is None and try_count > 0:
            if try_count != 20:
                sleep(10)
            self.logger.debug('attempting to aquire lock')
            message = DAQCoordCommand(command='acquire lock')
            self.socket.send(message.serialize())
            response = DAQCoordResponse.parse(self.socket.recv(), self.logger)
            self.logger.debug(f'Received response: {response}')
            try_count -= 1
        if response.content is not None:
            self.logger.info('Acquired lock')
            self.lock = response.content
        else:
            raise ValueError('Lock could not be acquires within the timeout '
                             'period')

    def initialize(self, config: dict, start_tdcs: bool = False):
        message = DAQCoordCommand(command='initialize',
                                  locking_token=self.lock,
                                  config=config)
        self.logger.info('Sending initial config')
        self.logger.debug(f'\n{config}')
        self.socket.send(message.serialize())
        self.zmq_transaction_running = True
        self.raw_response = self.socket.recv()
        self.zmq_transaction_running = False
        response = DAQCoordResponse.parse(self.raw_response, self.logger)
        self.logger.debug(f'Received response:\n{response}')
        if response.type == 'error':
            self.logger.error(
                f"Received error from coordinator: {response.content}")
            raise ValueError(
                'Server responded with an error state: '
                f'{response.content}')
        self.logger.info('initialized')
        if start_tdcs:
            rocs = list(filter(
                lambda x: 'roc_s' in x, config['target'].keys()))
            self._start_tdcs(rocs)

    def measure(self, run_config: dict, readback=False):
        message = DAQCoordCommand(command='measure', locking_token=self.lock,
                                  config=run_config, readback=readback)
        raw_msg = message.serialize()
        self.socket.send(raw_msg)
        self.zmq_transaction_running = True
        self.raw_response = self.socket.recv()
        self.zmq_transaction_running = False
        response = DAQCoordResponse.parse(self.raw_response, self.logger)
        if response.type == 'error':
            self.logger.error(
                f"Received error from coordinator: {response.content}")
            raise ValueError(
                'Server responded with an error state: '
                f'{response.content}')
        return response.content

    def read_target(self, config: dict = None):
        message = DAQCoordCommand(
            command='read_target',
            locking_token=self.lock,
            config=config
        )
        raw_msg = message.serialize()
        self.socket.send(raw_msg)
        self.zmq_transaction_running = True
        self.raw_response = self.socket.recv()
        self.zmq_transaction_running = False
        response = DAQCoordResponse.parse(self.raw_response, self.logger)
        if response.type == 'error':
            self.logger.error(
                f"Received error from coordinator: {response.content}")
            raise ValueError(
                'Server responded with an error state: '
                f'{response.content}')
        return response.content

    def __del__(self):
        if self.zmq_transaction_running:
            response = self.socket.recv()
            response = DAQCoordResponse.parse(response, self.logger)
            if response.type != 'error':
                self.logger.info(
                    f"Receive message of type {response.type}"
                    "from coordinator, discarding")
            if response.type == 'error':
                self.logger.error(
                    f"Received error from coordinator: {response.content}")
        if self.lock is not None:
            message = DAQCoordCommand(
                command='release lock', locking_token=self.lock)
            self.socket.send(message.serialize())
            response = DAQCoordResponse.parse(self.socket.recv(), self.logger)
            if response.type == 'error':
                self.logger.error(
                    f"Received error from coordinator: {response.content}")
                raise ValueError(
                    'Server responded with an error state: '
                    f'{response.content}')
        self.socket.close()
        self.io_context.destroy()

    def reset_target(self):
        self.logger.info('Resetting Target')
        message = DAQCoordCommand(
            command='reset_target',
            locking_token=self.lock,
            config=None)
        self.socket.send(message.serialize())
        self.zmq_transaction_running = True
        self.raw_response = self.socket.recv()
        self.zmq_transaction_running = False
        response = DAQCoordResponse.parse(self.raw_response, self.logger)
        if response.type == 'error':
            self.logger.error(
                f"Received error from coordinator: {response.content}")
            raise ValueError(
                'Server responded with an error state: '
                f'{response.content}')
        elif response.type != 'ack':
            self.logger.error(
                "Received unexpected response from coordinator: "
                f"{response}")
            raise ValueError(
                'Server responded with an error state: '
                f'{response.content}')
        self.logger.info('Target successfully reset')
        return

    def _start_tdcs(self, rocs: list, readback=False):
        self.logger.info('Starting up Master TDCs')
        config = {'target': {}}
        for roc in rocs:
            config['target'][roc] = \
                {'MasterTdc':
                    {0: {'EN_MASTER_CTDC_VOUT_INIT': 1,
                         'VD_CTDC_P_DAC_EN': 1,
                         'VD_CTDC_P_D': 16,
                         'EN_MASTER_FTDC_VOUT_INIT': 1,
                         'VD_FTDC_P_DAC_EN': 1,
                         'VD_FTDC_P_D': 16},
                     1: {'EN_MASTER_CTDC_VOUT_INIT': 1,
                         'VD_CTDC_P_DAC_EN': 1,
                         'VD_CTDC_P_D': 16,
                         'EN_MASTER_FTDC_VOUT_INIT': 1,
                         'VD_FTDC_P_DAC_EN': 1,
                         'VD_FTDC_P_D': 16}
                     }
                 }
        message = DAQCoordCommand(command='configure',
                                  config=config,
                                  locking_token=self.lock,
                                  readback=readback)
        self.logger.info('Sending configure command with config: ')
        self.logger.debug(f'\n{config}')
        self.socket.send(message.serialize())
        self.zmq_transaction_running = True
        self.raw_response = self.socket.recv()
        self.zmq_transaction_running = False
        response = DAQCoordResponse.parse(self.raw_response, self.logger)
        if response.type == 'error':
            self.logger.error(
                f"Received error from coordinator: {response.content}")
            raise ValueError(
                'Server responded with an error state: '
                f'{response.content}')

        config = {'target': {}}
        for roc in rocs:
            config['target'][roc] = \
                {'MasterTdc':
                    {0: {'EN_MASTER_CTDC_VOUT_INIT': 0,
                         'EN_MASTER_FTDC_VOUT_INIT': 0
                         },
                     1: {'EN_MASTER_CTDC_VOUT_INIT': 0,
                         'EN_MASTER_FTDC_VOUT_INIT': 0
                         }
                     }
                 }
        message = DAQCoordCommand(command='configure',
                                  config=config,
                                  locking_token=self.lock,
                                  readback=readback)
        self.logger.info('Sending configure command with config: ')
        self.logger.debug(f'\n{config}')
        self.socket.send(message.serialize())
        self.zmq_transaction_running = True
        self.raw_response = self.socket.recv()
        self.zmq_transaction_running = False
        self.logger.info('Master TDCs set up')


class DAQCoordinator():

    def __init__(self, network_config, capture_data):
        self.capture_data = capture_data
        self.network_config = network_config
        self.logger = logging.getLogger('daq_coordinator')
        self.lock = None
        self.initialized = False
        self.run_counter = 0
        self.measurement_data_path = Path('/tmp/measurement_data.raw')
        self.logger.info('creating ZMQ context and socket for Coordinator')
        self.io_context = zmq.Context()
        self.command_socket = self.io_context.socket(zmq.REP)
        socket_address = \
            f"tcp://{self.network_config['daq_coordinator']['hostname']}:" +\
            f"{self.network_config['daq_coordinator']['port']}"
        self.command_socket.bind(socket_address)
        self.logger.debug('bound to: %s' % socket_address)
        self.logger.info('Initializing target-adapter')
        self.target = SCClient(
            self.network_config['server']['hostname'],
            self.network_config['server']['sc_ctrl_port'])
        self.logger.info('Initializing DAQ-System adapter')
        self.daq_system = DAQSystem(
            client_hostname=network_config['client']['hostname'],
            client_port=network_config['client']['port'],
            server_hostname=network_config['server']['hostname'],
            server_port=network_config['server']['daq_ctrl_port'],
            data_port=network_config['server']['data_port'])

    def run(self):
        while True:
            message = self.command_socket.recv()
            daq_response = None
            try:
                daq_command = DAQCoordCommand.parse(message)
                command = daq_command.command
                self.logger.info(f'Received a {command} command')
            except DAQError as e:
                self.logger.info('Received invalid command')
                daq_response = DAQCoordResponse(
                    type='error',
                    content=f"Parsing of the daq_command "
                            f"{daq_command.command} failed with "
                            f"the error: {e}")
                self.command_socket.send(daq_response.serialize(self.logger))
                self.lock = None
                continue

            # check for the different kind of messages
            if command == 'acquire lock':
                if self.lock is None:
                    self.lock = str(uuid.uuid1())
                    self.logger.info(
                        f'No lock set, acquiring lock: {self.lock}')
                    daq_response = DAQCoordResponse(type='lock',
                                                    content=self.lock)
                else:
                    self.logger.info(
                        'DAQ-System is locked, sending empty response')
                    daq_response = DAQCoordResponse(type='lock',
                                                    content=None)
                self.command_socket.send(daq_response.serialize(self.logger))
                continue

            # all messages beyond this point need a locking token to work
            # so respond with an error here if no locking token can be found
            msg_lock = daq_command.locking_token
            self.logger.debug(f'Received lock from cliend: {msg_lock}')
            if self.lock != msg_lock:
                self.logger.warn(
                    f'Lock in the message does not match {self.lock}'
                    'responing with access denied')
                daq_response = DAQCoordResponse(type='access denied')
                self.command_socket.send(daq_response.serialize(self.logger))
                continue

            if command == 'release lock':
                daq_response = DAQCoordResponse(
                    type='ack',
                    content=None)
                self.lock = None
                self.logger.info('releasing lock')
                self.command_socket.send(daq_response.serialize(self.logger))
                if self.initialized:
                    self.logger.info('Deinitializing DAQ system')
                    self.daq_system.tear_down_data_taking_context()
                    self.initialized = False
                    self.run_counter = 0
                continue

            if command == 'initialize':
                config = daq_command.config
                self.logger.debug(
                    'Received initialization config:\n' +
                    yaml.dump(config))
                try:
                    self.logger.info('Resetting ROCs')
                    self.target.reset()
                    self.logger.debug(
                        'Initializing DAQ-System and setting up data'
                        'taking context')
                    self.daq_system.initialize(initial_config=config)
                    self.daq_system.setup_data_taking_context()
                except ValueError as err:
                    self.logger.warn(
                        'initialization of the daq system failed, received'
                        f'error: {err.args[0]}')
                    error_msg = 'During initialization of the daq system ' + \
                                f'an error ocurred: {err.args[0]}'
                    daq_response = DAQCoordResponse(
                        type='error',
                        content=error_msg
                    )
                    self.command_socket.send(
                        daq_response.serialize(self.logger))
                    continue
                try:
                    target_config = config['target']
                except KeyError:
                    target_config = {}
                if target_config != {}:
                    self.logger.debug('Initializing target system')
                    try:
                        self.logger.debug(
                            'Initializing Rocs\n'
                            f'{yaml.dump(target_config)}')
                        self.target.set(target_config,
                                        readback=daq_command.readback)
                    except ValueError as err:
                        self.logger.warn(
                            'target configuration failed. Got the error: '
                            f'{err.args[0]} from the sc-server')
                        error_msg = \
                            'During configuration of the ROCs an error ' + \
                            f'ocurred: {err.args[0]}'
                        daq_response = DAQCoordResponse(
                            type='error',
                            content=error_msg
                        )
                        self.command_socket.send(
                            daq_response.serialize(self.logger))
                        continue
                else:
                    self.logger.debug('No target config found, '
                                      'target initialization skipped')
                self.logger.info('DAQ-System and target initialized')
                response = DAQCoordResponse(type='ack')
                self.initialized = True
                self.command_socket.send(response.serialize(self.logger))
                continue

            if command == 'configure':
                self.logger.info('Configuring DAQ-System')
                config = daq_command.config
                self.logger.debug(
                    'Received system config:\n' +
                    yaml.dump(config))
                try:
                    self.daq_system.configure(config)
                except ValueError as err:
                    self.logger.warn(
                        'initialization of the daq system failed, received'
                        f'error: {err.args[0]}')
                    error_msg = 'During initialization of the daq system ' + \
                                f'an error ocurred: {err.args[0]}'
                    daq_response = DAQCoordResponse(
                        type='error',
                        content=error_msg
                    )
                    self.command_socket.send(
                        daq_response.serialize(self.logger))
                    continue
                try:
                    target_config = config['target']
                except KeyError:
                    target_config = {}
                if target_config != {}:
                    self.logger.debug('Configuring target system')
                    try:
                        self.logger.debug(
                            'Configuring Rocs\n'
                            f'{yaml.dump(target_config)}')
                        self.target.set(target_config,
                                        readback=daq_command.readback)
                    except ValueError as err:
                        self.logger.warn(
                            'target configuration failed. Got the error: '
                            f'{err.args[0]} from the sc-server')
                        error_msg = \
                            'During configuration of the ROCs an error ' + \
                            f'ocurred: {err.args[0]}'
                        daq_response = DAQCoordResponse(
                            type='error',
                            content=error_msg
                        )
                        self.command_socket.send(
                            daq_response.serialize(self.logger))
                        continue
                else:
                    self.logger.debug('No target config found, '
                                      'target initialization skipped')
                self.logger.info('DAQ-System and target configured')
                response = DAQCoordResponse(type='ack')
                self.command_socket.send(response.serialize(self.logger))
                continue

            if command == 'measure':
                if not self.initialized:
                    self.logger.error(
                        'DAQ system has not been initialized yet')
                    daq_response = DAQCoordResponse(
                        type='error',
                        content='The daq coordinator must be initialized'
                        ' before measurements can be taken'
                    )
                    self.command_socket.send(
                        daq_response.serialize(self.logger))
                    continue
                config = daq_command.config
                self.logger.debug(
                    'Received run configuration:\n'
                    + yaml.dump(config))
                try:
                    target_config = config['target']
                except KeyError:
                    target_config = {}
                if target_config != {}:
                    self.logger.info('Configuring target system')
                    try:
                        self.target.set(target_config,
                                        readback=daq_command.readback)
                    except ValueError as err:
                        self.logger.warn(
                            'target configuration failed. Got the error: '
                            f'{err.args[0]} from the sc-server')
                        error_msg = \
                            'During configuration of the ROCs an error ' + \
                            f'ocurred an error ocurred: {err.args[0]}'
                        daq_response = DAQCoordResponse(
                            type='error',
                            content=error_msg
                        )
                        self.command_socket.send(
                            daq_response.serialize(self.logger))
                        continue
                else:
                    self.logger.debug('No target config found, '
                                      'target configuration skipped')
                try:
                    self.logger.debug('Configuring daq-system')
                    self.daq_system.configure(config)
                    self.logger.info('Acquiring Data')
                    self.daq_system.take_data(self.measurement_data_path)
                except ValueError as err:
                    self.logger.warn(
                        'Data taking failed. Got the error: '
                        f'{err.args[0]} from the DAQ-system')
                    error_msg = \
                        'During the Data taking the DAQ system encountered' \
                        f' an error: {err.args[0]}'
                    daq_response = DAQCoordResponse(
                        type='error',
                        content=error_msg
                    )
                    self.command_socket.send(
                        daq_response.serialize(self.logger))
                    continue
                except RuntimeWarning as warn:
                    self.logger.warn(
                        'Data taking failed. Got the warning: '
                        f'{warn.args[0]} from the DAQ-system')
                    error_msg = \
                        'The DAQ system encountered' \
                        f' an error: {warn.args[0]}'
                    daq_response = DAQCoordResponse(
                        type='error',
                        content=error_msg
                    )
                    self.command_socket.send(
                        daq_response.serialize(self.logger))
                with open(self.measurement_data_path, 'rb') as data_file:
                    self.logger.info('Sending Acquired data to clinet')
                    daq_response = DAQCoordResponse(
                        type='data',
                        content=data_file.read()
                    )
                if self.capture_data:
                    storage_file_name = \
                        Path(self.lock + f'_run_{self.run_counter}_data.raw')
                    self.logger.info(
                        'Storing Data locally in file: '
                        f'{storage_file_name}')
                    os.rename(self.measurement_data_path,
                              storage_file_name.absolute())
                self.run_counter += 1
                self.command_socket.send(daq_response.serialize(self.logger))
                self.logger.info('Data sent')
                continue

            if command == 'read_target':
                read_config = daq_command.config
                self.logger.info('Received request to read out the target')
                try:
                    config = self.target.get_from_hardware(read_config)
                    self.logger.info(
                        'Read Config from the hardware, sending it to '
                        'the client')
                    daq_response = DAQCoordResponse(
                        type='target_config',
                        content={'target': config}
                    )
                except ValueError as err:
                    self.logger.error(
                        f'Received error from the Target: {err.args[0]}')
                    daq_response = DAQCoordResponse(
                        type='error',
                        content=f'{err.args[0]}'
                    )
                self.command_socket.send(daq_response.serialize(self.logger))
                self.logger.info('target config sent to client')
                continue

            if command == 'reset_target':
                self.logger.info('Received request to reset target')
                try:
                    self.target.reset()
                    self.logger.info('Sending target reset command')
                    daq_response = DAQCoordResponse(
                        type='ack')
                except ValueError as err:
                    self.logger.error(
                        f'Received error from the Target: {err.args[0]}')
                    daq_response = DAQCoordResponse(
                        type='error',
                        content=f'{err.args[0]}'
                    )
                self.command_socket.send(daq_response.serialize(self.logger))
                self.logger.info('Target Reset')
                continue

            if command == 'shutdown':
                daq_response = DAQCoordResponse(
                    type='ack'
                )
                self.command_socket.send(daq_response.serialize())
                self.logger.info('Shutting down')
                break


_log_level_dict = {'DEBUG': logging.DEBUG,
                   'INFO': logging.INFO,
                   'WARNING': logging.WARNING,
                   'ERROR': logging.ERROR,
                   'CRITICAL': logging.CRITICAL}


@click.command()
@click.argument('netcfg', type=click.File('r'),
                metavar='[network configuration file]')
@click.option('--log/--no-log', type=bool, default=True,
              help='Enable logging and append logs to the filename passed to '
                   'this option')
@click.option('--loglevel', default='INFO',
              type=click.Choice(['DEBUG', 'INFO',
                                 'WARNING', 'ERROR', 'CRITICAL'],
                                case_sensitive=False))
@click.option('--client_output', default=None,
              help='file to place the output of the client into, by default'
                   ' the output of the daq-client is not captured')
@click.option('--capture_data/--dont_capture_data', default=False)
def main(netcfg, log, loglevel, client_output, capture_data):
    if client_output is not None:
        client_output = Path(client_output)
    if log:
        logging.basicConfig(filename='daq-coord.log',
                            filemode='a+',
                            level=_log_level_dict[loglevel],
                            format='[%(asctime)s] %(levelname)-10s:'
                                   '%(name)-50s %(message)s')
    logging.info('Read in network config')
    try:
        netcfg = yaml.safe_load(netcfg.read())
        netcfg['client'] = {}
        netcfg['client']['hostname'] = 'localhost'
        netcfg['client']['port'] = 6001
    except yaml.YAMLError as err:
        sys.exit('Error reading in the network config:\n' +
                 + str(err) + '\nexiting ..')

    logging.info('Starting daq-client')
    daq_client_path = shutil.which('daq-client')
    if daq_client_path is None:
        logging.error('daq-client executable not found, exiting')
        print('daq-client executable not found, exiting')
        sys.exit(1)
    if client_output is None:
        logging.info('discarding client output')
        client_process_out = sp.DEVNULL
    else:
        logging.info(f'Writing client output into {client_output.absolute()}')
        client_process_out = open(client_output, 'a+')
        client_process_out.write(f'Started daq-client on {ctime()}')
    client_process = sp.Popen(
        [daq_client_path, '-p', str(netcfg['client']['port'])],
        stdout=client_process_out)
    daq_coordinator = DAQCoordinator(netcfg, capture_data)
    logging.info('DAQ-Coordinator initialized')
    try:
        daq_coordinator.run()
    except KeyboardInterrupt:
        client_process.kill()
        if client_output is not None:
            client_process_out.close()
        sys.exit(1)
    client_process.kill()
    if client_output is not None:
        client_process_out.close()



"""
Module containing the classes that together constitute a measurement
and encapsulate the different steps needed to take a measurement using the
hexaboard
"""
from pathlib import Path
from functools import reduce
from multiprocessing import Process, Queue
import os
from time import sleep
import operator
import shutil
import luigi
import yaml
import zmq
from uproot.exceptions import KeyInFileError
from . import dict_utils as dtu
from . import config_utilities as cfu
from . import analysis_utilities as anu
from .daq_coordination import DAQCoordCommand, DAQCoordResponse
from .config_errors import DAQError
from luigi.parameter import ParameterVisibility
from luigi import Nop


class Well(luigi.Task):
    """
    A Scan over one parameter or over other scans

    The scan uses the base configuration as the state of the system
    and then modifies it by applying patches constructed from
    parameter/value pairs passed to the scan and then calling either
    the measurement task or a sub-scan with the patched configurations
    as their respective base configurations
    """
    # parameters describing the position of the parameters in the task
    # tree
    system_state = luigi.DictParameter(significant=True,
                                       visibility=ParameterVisibility.PRIVATE)
    id = luigi.IntParameter(significant=True)
    priority = luigi.OptionalParameter(significant=False, default=0)

    def requires(self):
        """
        Determine the measurements that are required for this scan to proceed.

        The Scan class is a recursive task. For every parameter(dimension) that
        is specified by the parameters argument, the scan task requires a
        set of further scans, one per value of the values entry associated with
        the parameter that the current scan is to scan over, essentially
        creating the Cartesian product of all parameters specified.
        """
        from .valve_yard import ValveYard
        if self.system_state['procedure']['calibration'] is not None:
            return ValveYard(system_state=self.system_state,
                             task_name=self.system_state['procedure']['calibration'],
                             priority=self.priority)

    def output(self):
        """
        generate the output file for the scan task

        If we are in the situation of being called by the Fracker
        (first if condition) it is the job of the DataField to simply produce
        the raw files. It then also needs to figure out what files still need
        to be generated, as such it needs check what files have already been
        converted by the fracker. The fracker will fail and stall the rest of
        the luigi pipeline if it can't unpack the file. The user then needs to
        rerun the datenraffinerie
        """
        name = self.system_state['procedure']['name']
        output_dir = Path(self.system_state['output_path'])
        output = {}
        full_calib_path = output_dir / 'full-calibration.yaml'
        output['full-calibration'] = luigi.LocalTarget(
                full_calib_path.resolve())
        run_ids = reduce(operator.mul,
                         map(lambda x: len(x['values']),
                             self.system_state['procedure']['parameters']))
        output['default'] = luigi.LocalTarget(output_dir /
                                              name + '_full_init.yaml')
        output['data'] = []
        output['raw'] = []
        output['config'] = []
        for i in run_ids:
            output['raw'].append(
                    luigi.LocalTarget(output_dir / name + f'_{i}.raw',
                                      format=Nop))
            output['config'].append(
                    luigi.LocalTarget(output_dir / name + f'_{i}.yaml'))
            if not self.system_state['procedure']['merge']:
                output['data'].append(
                        luigi.LocalTarget(output_dir / name + f'_{i}.h5'))
            else:
                output['data'].append(
                        luigi.LocalTarget(output_dir / name + '_merged.h5'))
        return output

    def run(self):
        """
        concatenate the files of a measurement together into a single file
        and write the merged data, or if the 'loop' parameter is set, it
        performs the measurements and lets the fracker handle the initial
        conversion into usable files if loop is set the fracker also does
        the merging at the end so in that case it is really 'just' there
        to acquire the data'.
        """

        # the fracker required us so we acquire the data and don't do any
        # further processing
        self.system_state = cfu.unfreeze(self.system_state)
        network_config = self.system_state['network_config']

        # open the connection to the daq coordinator
        context = zmq.Context()
        socket = context.socket(zmq.REQ)
        socket.connect(
            f"tcp://{network_config['daq_coordinator']['hostname']}:"
            f"{network_config['daq_coordinator']['port']}"
        )
        # load the configurations
        try:
            with self.input()['full-calibration'].open('r') as calibf:
                calibration = yaml.safe_load(calibf.read())
        except KeyError:
            calibration = {}

        # generate the default config
        default_config = cfu.generate_system_default_config(
                self.system_state['procedure'])

        # generate initial config and apply calibration
        initial_config = cfu.generate_init_config(
                self.system_state['procedure'])
        dtu.update_dict(initial_config, calibration, in_place=True)

        # write

        # generate the patches for the daq procedure
        daq_patches = cfu.generate_patches(self.system_state)

        # set up the fracker subprocesses
        fracker_task_queue = Queue()
        fracker_processes = []
        for i in range(self.system_state['workers']):
            p = Process(target=self.frack, args=(fracker_task_queue,
                                                 default_config))
            fracker_processes.append(p)

        # aquire lock from the daq coordinator
        lock_key = None
        retries = 100
        while True:
            command = DAQCoordCommand(command='acquire lock')
            socket.send(command.serialize())
            message = socket.recv()
            response = DAQCoordResponse.parse(message)
            if response['type'] == 'lock':
                if response['content'] is None:
                    sleep(.5)
                    retries -= 1
                    if retries > 0:
                        continue
                    else:
                        socket.close()
                        context.term()
                        raise DAQError(
                            "Failed to aquire Lock in procedure: "
                            f"{self.system_state['procedure']['name']}")
                else:
                    lock_key = response['content']
                    break
            elif response['type'] != 'lock':
                raise DAQError(
                        "Failed to aquire lock, received unexpected response"
                        f"{response}"
                        )

        # ------------------ lock aquired ---------------------------

        # load default config
        command = DAQCoordCommand(
                command='load defaults',
                locking_token=lock_key,
                config=default_config
        )
        socket.send(command.serialize())
        resp = socket.recv()
        resp = DAQCoordResponse.parse(resp)
        if resp['type'] != 'ack':
            raise DAQError(f"received unexpected repsonse: {resp}")

        # compute the filenames for the runs
        output_files = self.output()
        for patch, raw_file, config_file in\
                zip(daq_patches, output_files['raw'], output_files['config']):
            # build run config and write it to disk
            if raw_file.exists():
                continue
            run_config = dtu.update_dict(initial_config, patch)
            serialized_config = yaml.safe_dump(run_config)
            with config_file.open('w') as cf:
                cf.write(serialized_config)

            # send measure command to daq coordinator
            command = DAQCoordCommand(
                    type='measure',
                    config=run_config,
                    locking_token=lock_key
            )
            socket.send(command.serialize())
            message = socket.recv()
            response = DAQCoordResponse.parse(message)
            if response['type'] == 'data':
                with raw_file.open('w') as raw_data_file:
                    raw_data_file.write(response['data'])

            # put the fracking tasks on the queue
            columns = self.system_state['procedure']['data_columns']
            mode = self.system_state['procedure']['mode']
            command = 'format'
            fracker_task_queue.put((command, raw_file.path,
                                    config_file.path,
                                    columns, mode))

        # stop fracking fracker_processes
        for p in fracker_processes:
            fracker_task_queue.put(('stop', '', '', '', ''))
        for p in fracker_processes:
            p.join()

        # release the lock
        command = DAQCoordCommand(type='release lock',
                                  locking_token=lock_key)
        socket.send(command.serialize())
        resp = DAQCoordResponse.parse(socket.recv())
        if resp['type'] != 'ack':
            raise DAQError(f'Error from the daq coordinator: {resp}')

        # ------------------------- lock released -------------------------

        # join all previous tasks
        # merge the files if the procedure is configured to do so
        if self.system_state['merge']:
            raw_files = [r.path for r in self.output()['raw']]
            fracked_file_paths = [
                    Path(os.path.splitext(rf)[0] + '.h5').absolute()
                    for rf in raw_files
            ]
            for p in fracked_file_paths:
                if not Path(p).exists():
                    base_name = os.path.splitext(p.resolve())[0]
                    os.remove(base_name + '.raw')
                    raise DAQError(f'Fracker subtask failed, '
                                   f'{base_name + "h5"} could not be found')

            if len(fracked_file_paths) == 1:
                shutil.copy(fracked_file_paths[0],
                            self.output()['data'].path)
                return
            result = anu.run_turbo_pump(self.output()['data'].path,
                                        fracked_file_paths)
            if result != 0:
                raise DAQError("turbo-pump crashed")

    def frack_data(raw_data_queue, default_config):
        # TODO implement queue logic
        command, raw_file_path, config_file_path, columns, mode =\
                raw_data_queue.get()
        while command != "stop":
            unpacked_file_path = os.path.splitext(raw_file_path)[0] + '.root'
            formatted_data_path = os.path.splitext(raw_file_path)[0] + '.h5'
            result = anu.unpack_raw_data_into_root(
                    raw_file_path,
                    unpacked_file_path,
            )
            # if the unpaack command failed remove the raw file to
            # trigger the Datafield to rerun the data taking
            if result != 0 and unpacked_file_path.exists():
                os.remove(unpacked_file_path)
                os.remove(raw_file_path)
                continue
            if not unpacked_file_path.exists():
                os.remove(raw_file_path)
                continue
            with open(config_file_path, 'r') as rc:
                run_config = yaml.safe_load(rc.read())
            complete_config = dtu.update_dict(default_config, run_config),
            # if the fracker can be found run it
            if shutil.which('fracker') is not None:
                retval = anu.run_compiled_fracker(
                        str(unpacked_file_path.absolute()),
                        str(formatted_data_path.absolute()),
                        complete_config,
                        mode,
                        columns)
                if retval != 0:
                    print("The fracker failed!!!")
                    if formatted_data_path.exists():
                        os.remove(formatted_data_path)
            # otherwise fall back to the python code
            else:
                try:
                    anu.reformat_data(unpacked_file_path,
                                      formatted_data_path,
                                      complete_config,
                                      mode,
                                      columns)
                except KeyInFileError:
                    os.remove(unpacked_file_path)
                    os.remove(raw_file_path)
                    continue
                except FileNotFoundError:
                    continue
            os.remove(unpacked_file_path)
            command, raw_file_path, config_file_path, columns, mode =\
                raw_data_queue.get()



"""
The ValveYard is a class in the datenraffinerie that is responsible
for parsing the user configuration and requiring the right scan or
analysis task and passing it all necessary values from their proper
functioning

Author: Alexander Becker (a.becker@cern.ch)
Date: 2021-12-16
"""
import os
from pathlib import Path
from copy import deepcopy
import luigi
import yaml
import logging
from .scan import DataField
from .distillery import Distillery
from . import config_utilities as cfu
from . import dict_utils as dtu
from . import config_validators as cfv
from . import control_adapter as ctrl
from . import vlvyrd_utils as vyu


module_logger = logging.getLogger(__name__)


class ValveYard(luigi.Task):
    system_state = luigi.DictParameter(significant=True)
    task_name = luigi.Parameter(significant=True)
    priority = luigi.OptionalParameter(significant=False, default=0)

    def output(self):
        """
        determin the files that are to be returned by the ValveYard to them
        calling tasks

        if the tasks are run in event mode, the output files
        are not concatenated but are instead just passed along to the
        calling task

        as this is the first method called during execution of the ValveYard,
        also determin if the
        """

        output_directory = Path(self.system_state['output_path'])
        input = self.input()
        output = {}
        output['data'] = [i['data'] for i in input]
        output['full-calibration'] = Path(output_directory) /\
            str(self.task_name) + '-full-calibration.yaml'
        return output

    def requires(self):
        """ A wrapper that parses the configuration and starts the procedure
        with the corresponding procedure label

        :raises: ConfigFormatError if either the type of the configuration
            entry
            does not match the allowed ones or if the YAML is malformed
        :raises: DAQConfigError if the DAQ configuration is malformed
        """
        self.system_state = cfu.unfreeze(self.system_state)
        root_cfg_file = self.system_state['root_config_path']
        cfv.current_path = Path(root_cfg_file)
        configuration = cfv.main_config.validate(root_cfg_file)
        output_directory = Path(self.system_state['output_path'])

        available_procedures = configuration['procedures']\
            + configuration['libraries']
        available_workflows = configuration['workflows']

        try:
            procedures = vyu.get_procedure_with_name(
                    self.task_name,
                    procedures=available_procedures,
                    workflows=available_workflows)
        except ctrl.DAQConfigError:
            raise ctrl.DAQConfigError(
                    f'No procedure with name: {self.task_name} found')

        dependencies = []
        for i, procedure in enumerate(procedures):
            output_dir = output_directory / procedure['name']
            if not output_directory.exists():
                os.makedirs(output_directory)
            if not output_dir.exists():
                os.makedirs(output_dir)
            task_state = deepcopy(self.system_state)
            task_state['procedure'] = procedure
            dependency_level = self.system_state['dependency_level']
            if procedure['type'] == 'daq':
                dependencies.append(
                    DataField(system_state=task_state,
                              id=self.id,
                              priority=len(procedures) * i * 100 * dependency_level))
            if procedure['type'] == 'analysis':
                dependencies.append(
                    Distillery(system_state=task_state,
                               id=self.id,
                               priority=i * 100 * dependency_level))
        return dependencies

    def run(self):
        in_calib_files = []
        ofiles = self.output()
        for ifiles in self.input():
            in_calib_files.append(ifiles['full-calibration'])
        full_calibration = {}
        for calibfile in in_calib_files:
            with calibfile.open('r') as cf:
                calib = cfu.load_configuration(cf)
                dtu.update_dict(full_calibration, calib, in_place=True)
        with ofiles['full-calibration'].open() as fcf:
            fcf.write(yaml.safe_dump(full_calibration))



from schema import Schema, Or
from os import path
from . import errors
from multiprocessing import Queue, Process
from .analysis_utilities import daq_columns, data_columns, event_mode_data_columns


class Frackers():
    modes = ['full', 'summary']
    fracker_command_schema = Schema({
            'command': Or('reformat', 'stop'),
            'data': Or(path.exists, None),
            'config': Or(path.exists, None)
        })

    def __init__(self, count, default_config_path, columns, mode):
        if not path.exists(default_config_path):
            raise errors.DAQError(f'{default_config_path} does not exist')
        if mode not in self.modes:
            raise errors.DAQError(f'{mode} not in {self.modes}')
        if mode == 'full':
            f_data_columns = event_mode_data_columns
        else:
            f_data_columns = data_columns
        for column in columns:
            if column not in f_data_columns or column not in 
        fracker_processes = []
        task_queue = Queue()
        for i in range(count):
            p = Process(
                    target=self.frack,
                    args=(task_queue, default_config_path, columns))
            fracker_processes.append(p)

    def frack(task_queue, default_config_path, columns):
        



import click
from . import dict_utils as dctu
import yaml


@click.group("cli")
@click.pass_context
def cli(ctx):
    if ctx.invoked_subcommand is None:
        click.echo('No command was given to the tool. Available commands are:'
                   '\n\tupdate\n\tdiff')


@cli.command()
@click.argument('original', type=click.File('r'))
@click.argument('patch', type=click.File('r'))
@click.option('--output', '-o', type=click.Path(), default=None,
              help='File that the updated dict is written to. '
                   'If no file is given the output is printed on stdout')
def update(original, patch, output):
    try:
        orig_dict = yaml.safe_load(original.read())
    except yaml.ScannerError:
        click.echo('The original file is not valid yaml syntax')
        exit(1)
    try:
        update = yaml.safe_load(patch.read())
    except yaml.ScannerError:
        click.echo('The patch file is not valid yaml syntax')
        exit(1)
    output_dict = dctu.update_dict(orig_dict, update)
    if output_dict is None:
        output_dict = {}
    if output is not None:
        with open(output, 'w+') as opf:
            opf.write(yaml.safe_dump(output_dict))
    else:
        print(yaml.safe_dump(output_dict))


@cli.command()
@click.argument('original', type=click.File('r'))
@click.argument('changed', type=click.File('r'))
@click.option('--output', '-o', type=click.Path(), default=None,
              help='File that the diff is written to. '
                   'If no file is given the output is printed on stdout')
def diff(original, changed, output):
    try:
        orig_dict = yaml.safe_load(original.read())
    except yaml.ScannerError:
        click.echo('The original file is not valid yaml syntax')
        exit(1)
    try:
        changed_dict = yaml.safe_load(changed.read())
    except yaml.ScannerError:
        click.echo('The patch file is not valid yaml syntax')
        exit(1)
    diff_dict = dctu.diff_dict(orig_dict, changed_dict)
    if diff_dict is None:
        diff_dict = {}
    if output is not None:
        with open(output, 'w+') as opf:
            opf.write(yaml.safe_dump(diff_dict))
    else:
        print(yaml.safe_dump(diff_dict))


if __name__ == '__main__':
    cli()



"""
Module containing the adapters to the daq system consisting of the
zmq-server, zmq-client and the zmq_i2c-server. This module is and
should be the only point of interaction between the Datenraffinerie
and the daq services provided by the above mentioned programs, as
such it encapsulates the peculiarities of the underlying DAQ system
and provides a uniform API for the Datenraffinerie to use.

This file needs to be adapted if the underlying DAQ programs change
their behaviour
"""
import os
import logging
import shutil
from pathlib import Path
import uuid
import zmq
import yaml
from time import sleep
from .dict_utils import diff_dict, update_dict

module_logger = logging.getLogger(__name__)


def _filter_out_network_config(config):
    """
    As there is minimal network configuration inside the daq system config
    we need to filter this out to be able to pass along the parameters that
    are intended for the actual server and client

    Also has to handle data weirdly because of technical debt in the daq c++
    software
    """
    # this weird contraption needs to be build because the current zmq server
    # and client expect the ENTIRE configuration (including hexaboard and every
    # other component to be sent to them
    out_config = {}
    hostname = None
    port = None
    for key, value in config.items():
        if 'hostname' == key:
            hostname = value
        elif 'port' == key:
            port = value
        else:
            out_config[key] = value
    return out_config, hostname, port


class DAQAdapter():
    """
    A representation of the DAQ side of the system. It encapsulates the
    zmq-server and zmq-client
    """
    variant_key_map = {'server': 'daq', 'client': 'client'}

    def __init__(self, variant: str, hostname: str,
                 port: int, data_port: int):
        # set up the logger for this object
        self.logger = logging.getLogger(
                __name__+f'.{variant}')

        # initialize the connection to the target server
        self.hostname = hostname
        self.ctrl_port = port
        self.data_port = data_port
        self.context = zmq.Context()
        self.logger.info(f'Initializing connection to {variant}')
        self.socket = self.context.socket(zmq.REQ)
        socket_url = f'tcp://{self.hostname}:{self.ctrl_port}'
        self.logger.debug(f'connecting to {socket_url}')
        self.socket.connect(socket_url)
        self.variant = variant
        self.running = False

        # initialize the configuration properly
        self.configuration = {self.variant_key_map[self.variant]: {}}

    def __del__(self):
        self.socket.close()
        self.context.destroy()

    def _clean_configuration(self, config):
        config, _, _ = \
                _filter_out_network_config(config)
        try:
            config = config[self.variant_key_map[self.variant]]
        except KeyError:
            try:
                config = config[self.variant]
            except KeyError:
                config = {}
        config = {self.variant_key_map[self.variant]: config}
        return config

    def update_config(self, config):
        update_dict(self.configuration,
                    self._clean_configuration(config),
                    in_place=True)

    def initialize(self, initial_config: dict):
        self.logger.debug("initializing")
        self.configuration = self._clean_configuration(initial_config)
        # include the data port in the configuration
        self.configuration[self.variant_key_map[self.variant]]['zmqPushPull_port'] \
            = self.data_port
        self.socket.send_string("initialize", zmq.SNDMORE)
        config_string = yaml.dump(self.configuration)
        self.logger.debug("sending config:\n" + config_string)
        self.socket.send_string(config_string)
        rep = self.socket.recv()
        rep = rep.decode('utf-8')
        if rep != "initialized":
            self.logger.critical(
                    "Server not successfully initialized."
                    f" Received: {rep}")
            raise ValueError("Server not successfully initialized")
        else:
            self.logger.debug(f"Received reply: {rep}")

    def configure(self, config: dict = {}, cache=False):
        self.logger.info('Configuring')
        config = self._clean_configuration(config)
        self.logger.debug(f'Updating config with:\n{config}')
        if self.variant == 'client':
            write_config = config
        else:
            # constrain the configuration that should be updated
            # to the configuration in the active menu
            try:
                active_menu = config['daq']['active_menu']
            except KeyError:
                try:
                    active_menu = self.configuration['daq']['active_menu']
                except KeyError:
                    raise ValueError('No Active menu found')
            try:
                current_config = {
                    'daq':
                    {'active_menu': active_menu,
                     'menus': {
                        active_menu:
                        self.configuration['daq']['menus'][active_menu]}
                     }
                    }
            except KeyError:
                current_config = \
                    {'daq':
                     {'active_menu': active_menu,
                      'menus': {active_menu: {}}
                      }
                     }
            config = update_dict(current_config, config)
            # check that we are only configuring the stuff in the active_menu
            if any(map(lambda x: active_menu != x,
                       config['daq']['menus'].keys())):
                raise RuntimeWarning(
                    'Configuring menus that are not the'
                    'active menu, these parameters are going to be'
                    'ignored.')
            self.configuration = update_dict(self.configuration, config)
            if cache:
                write_config = config \
                    if diff_dict(current_config, config) is not None else \
                    {self.variant_key_map[self.variant]: {}}
            else:
                write_config = config

        # if the config is empty we don't have to write anything
        if write_config[self.variant_key_map[self.variant]] == {}:
            self.logger.debug("Configuration did not change, "
                              "skipping configuration of the backend")
            return
        self.logger.debug(f'Writing config:\n{config}')
        config_string = yaml.dump(write_config)
        self.logger.debug(f"Sending config:\n{config_string}")
        self.socket.send_string('configure', zmq.SNDMORE)
        self.socket.send_string(config_string)
        reply = self.socket.recv_string()
        if "configured" != reply:
            self.logger.error(
                    "The configuration cannot be "
                    f" written to {self.hostname}. The daq component"
                    f"responded with {reply}")
            raise ValueError(
                    "The configuration cannot be "
                    f" written to {self.hostname}. The daq component"
                    f"responded with {reply}")
        self.logger.info("Configuration successful")
        self.configuration = update_dict(self.configuration, write_config)

    def reset(self):
        self.logger.debug("resetting daq-server by closing and reopening "
                          "the connection")
        self.socket.close()
        self.context.destroy()
        self.context = zmq.Context()
        self.socket = self.context.socket(zmq.REQ)
        self.socket.connect("tcp://"+str(self.ip)+":"+str(self.port))
        self.configuration = {}

    def start(self):
        """
        Start the aquisition of the data on the server and client
        """
        self.logger.info("Starting data-taking")
        status = "none"
        self.logger.debug(f'Checking status of {self.variant}')
        self.socket.send_string("status")
        status = self.socket.recv_string()
        self.logger.debug(f'Current state: {status}')
        if status != "configured":
            raise ValueError("Server not ready to start data taking")
        retries = 50
        post_link_alignment_state = 'configured'
        while retries > 0:
            self.logger.debug(
                    f'Attempting to start backend, {retries} trys remaining')
            self.socket.send_string("start")
            status = self.socket.recv_string()
            if status == 'running':
                post_link_alignment_state = status
                break
            retries -= 1
        # if there where no trys left
        if post_link_alignment_state != 'running':
            self.logger.critical(
                    'Unable to start server, probably bad link alignment')
            raise ValueError(
                    'Unable to Start server, probably bad link alignment')
        self.logger.info(
            f"successfully started backend, current state: {status}")
        return status

    def stop(self):
        self.logger.info("Stopping data-taking")
        self.socket.send_string("stop")
        rep = self.socket.recv_string()
        self.logger.debug(f"Status received from Server: {rep}")
        if rep != "configured":
            raise ValueError("Server did not stop correctly")
        return

    def take_data(self):
        """
        check if the current aquisition is ongoing or not
        """
        self.logger.info('Starting data taking run')
        # attempt to start server. Server only switches into the
        # running state if the links are aligned.
        # so we query until we run out of tries or the server switches to
        # running
        self.logger.info('links aligned, taking data')
        status = self.start()
        self.logger.info('waiting for run to finish')
        while status == "running":
            self.socket.send_string("status")
            status = self.socket.recv_string()
        if status == "configured":
            self.logger.info('Data taking run completed')
            return
        else:
            self.logger.critical(
                'Invalid status from the DAQ server: {status}')
            raise ValueError('Invalid status from the DAQ server: {status}')


class DAQSystem:
    """
    A class that encapsulates the interactions
    with the DAQ-system (the zmq-[server|client])

    The class implements a small two-state state machine that only allows
    data-taking via the 'take_data' function after the 'start_run' function
    has been called. The data taking is stopped via the 'stop_run'
    function that
    """

    def __init__(self, server_hostname: str, server_port: int, data_port: int,
                 client_hostname: str, client_port: int):
        """
        initialise the daq system by initializing it's static components
        (and server)
        """
        # set up the server part of the daq system (zmq-server)
        self.logger = logging.getLogger(__name__+'.DAQSystem')
        self.client_started = False
        self.data_port = data_port
        self.server = DAQAdapter('server', server_hostname,
                                 server_port, data_port)
        self.client = DAQAdapter('client', client_hostname, client_port,
                                 self.data_port)
        # set up the client part of the daq system (zmq-client)
        # the wrapping with the global needs to be done so that the client
        # accepts the configuration

    def __del__(self):
        self.tear_down_data_taking_context()

    def initialize(self, initial_config: dict):
        initial_config['client']['serverIP'] = self.server.hostname
        self.client.initialize(initial_config)
        self.server.initialize(initial_config)

    def configure(self, daq_config: dict = None):
        """
        configure the daq system before starting a data-taking run.
        """
        self.client.configure(daq_config, cache=True)
        self.server.configure(daq_config, cache=True)

    def setup_data_taking_context(self):
        client_config = {}
        self.logger.debug('Setting up data taking context')
        self.procedure_uuid = uuid.uuid1().hex
        self.daq_data_folder = Path('/tmp') / self.procedure_uuid
        client_config['outputDirectory'] = str(self.daq_data_folder)
        client_config['run_type'] = "Datenraffinerie"
        if not os.path.isdir(self.daq_data_folder):
            os.mkdir(self.daq_data_folder)
        self.client.configure(
                {self.client.variant_key_map[self.client.variant]:
                 client_config})
        self.logger.debug('Updated client configuration to\n' +
                          yaml.safe_dump(self.client.configuration))
        active_menu = self.server.configuration['daq']['active_menu']
        config_transition_server_config = {}
        config_transition_server_config['daq'] = {}
        config_transition_server_config['daq']['active_menu'] = active_menu
        config_transition_server_config['daq']['menus'] = {}
        config_transition_server_config['daq']['menus'][active_menu] = \
            self.server.configuration['daq']['menus'][active_menu]
        self.server.configure(config_transition_server_config)

    def take_data(self, output_data_path):
        if not self.client_started:
            self.client.start()
            self.client_started = True
        try:
            self.server.take_data()
        except ValueError as err:
            self.client.stop()
            raise ValueError('Data taking failed') from err
        wait_cycles = 20
        data_files = os.listdir(self.daq_data_folder)
        while wait_cycles > 0 and len(data_files) == 0:
            data_files = os.listdir(self.daq_data_folder)
            sleep(.1)
            wait_cycles -= 1
        if len(data_files) != 1:
            raise ValueError(
                    "More than one file was found in the"
                    f" {self.daq_data_folder.resolve()} folder")
        data_file = data_files[0]
        shutil.move(self.daq_data_folder / data_file, output_data_path)

    def tear_down_data_taking_context(self):
        if self.client_started:
            self.client.stop()
            self.client_started = False
        try:
            if os.path.exists(self.daq_data_folder):
                for file in self.daq_data_folder.iterdir():
                    os.remove(file)
                os.rmdir(self.daq_data_folder)
        except AttributeError:
            return



class ConfigPatchError(Exception):
    def __init__(self, message):
        self.message = message


class ConfigFormatError(Exception):
    def __init__(self, message):
        self.message = message



"""
Module containing the utilities to parse and properly prepare the configuration
provided by the user

The functions here are testable by calling pytest config_utilities.py
The tests can also be used as a guide on how to use the functions and
what to expect of their behaviour
"""
from copy import deepcopy
from typing import Union
import yaml
import jinja2
import os
from .config_errors import ConfigFormatError
from . import dict_utils as dtu
from . import config_validators as cvd
import logging
from itertools import chain


def get_procedure_configs(main_config_file: str, procedure_name: str,
                          calibration: dict = None, diff: bool = False):
    cvd.set_current_path(os.path.dirname(main_config_file))
    with open(main_config_file, 'r') as cfp:
        config = yaml.safe_load(cfp.read())
    config = cvd.main_config.validate(config)
    available_procedures = config['procedures'] + \
        list(chain(*config['libraries']))
    try:
        procedure = list(filter(lambda x: x['name'] == procedure_name,
                                available_procedures))[0]
    except IndexError:
        all_procedure_names = list(map(lambda x: x['name'],
                                   available_procedures))
        logging.critical(f"The procedure with name: {procedure_name} "
                         "could not be found, Available procedures are: "
                         f"{all_procedure_names}")
        raise ValueError("Procedure could not be found",
                         procedure_name, all_procedure_names)
    return procedure, generate_configurations(procedure, calibration, diff)


def generate_configurations(procedure: dict, calibration: dict = None,
                            diff: bool = False):
    """
    Given a procedure entry loaded from the config file, generate
    the configurations necessary for every run of the procedure
    """
    system_default_config = generate_system_default_config(procedure)
    system_init_config = generate_init_config(procedure)
    scan_patches = generate_patches(procedure)
    if calibration is not None:
        dtu.update_dict(system_init_config, calibration, in_place=True)
    if diff:
        full_system_init_config = dtu.update_dict(system_default_config,
                                                  system_init_config)
        system_init_config = dtu.diff_dict(system_default_config,
                                           full_system_init_config)
        current_state = full_system_init_config.copy()

        def run_config_generator():
            for _ in range(procedure['repeat']):
                for patch in scan_patches:
                    run_config = dtu.update_dict(full_system_init_config,
                                                 patch)
                    run_config_diff = dtu.diff_dict(current_state, run_config)
                    if run_config_diff is not None:
                        dtu.update_dict(current_state,
                                        run_config_diff,
                                        in_place=True)
                        yield run_config_diff
                    else:
                        yield {}
    else:
        system_init_config = dtu.update_dict(system_default_config,
                                             system_init_config)

        def run_config_generator():
            for _ in range(procedure['repeat']):
                for patch in scan_patches:
                    yield dtu.update_dict(system_init_config, patch)
    return system_default_config, system_init_config, run_config_generator(), \
        len(scan_patches) * procedure['repeat']


def load_configuration(config_path):
    """
    load the configuration dictionary from a yaml file

    :raises: ConfigFormatError if the input cannot be parsed by the yaml
    parser
    """
    with open(config_path, 'r', encoding='utf-8') as config_file:
        try:
            return yaml.safe_load(config_file.read())
        except yaml.YAMLError as e:
            raise ConfigFormatError("unable to load yaml") from e


def generate_patch_from_key(keys: Union[list, str], value):
    """
    Given a list of keys and a value generate a nested dict
    with one level of nesting for every entry in the list
    if any of the list entries is itself a list create a dict for every
    element of the list. Make s
    """
    # if the key is in fact a template for a yaml file
    if isinstance(keys, str):
        patch = yaml.safe_load(jinja2.Template(keys).render(value=value))
        if len(patch.keys()) != 1 and \
                (patch.keys()[0] != 'target' or patch.keys()[0] != 'daq'):
            patch = {'target': patch}
        return patch

    # this is needed to work with luigi as it turns stuffinto
    # tuples
    keys = list(keys)
    # here we need to insert 'target' if the key 'target' or 'daq' is not
    # present as the top level key to extend the scannable parameters to both
    # the daq and target configuration
    # now the rest of the generation can run as usual
    if len(keys) == 0:
        return {}
    keys.reverse()
    current_root = value
    for key in keys:
        level = {}
        if isinstance(key, list) or isinstance(key, tuple):
            for subkey in key:
                level[subkey] = deepcopy(current_root)
        else:
            level[key] = current_root
        current_root = level
    return current_root


def build_dimension_patches(scan_dimension):
    scan_values = scan_dimension['values']
    patch_set = []
    default_set = []
    try:
        scan_dim_key = scan_dimension['key']
        for val in scan_values:
            patch = generate_patch_from_key(scan_dim_key, val)
            patch_set.append(patch)
            default_set.append({})
        return patch_set, default_set
    except KeyError:
        try:
            template = scan_dimension['template']
            default_template = scan_dimension['default']
            template = jinja2.Template(template)
            default_template = jinja2.Template(default_template)
            patch_set.append(
                yaml.safe_load(template.render(value=scan_values[0])))
            for prev_val, val in zip(scan_values[:-1], scan_values[1:]):
                default_set.append(
                    yaml.safe_load(default_template.render(value=prev_val)))
                patch_set.append(yaml.safe_load(template.render(value=val)))
            default_set.append(
                yaml.safe_load(default_template.render(value=scan_values[-1])))
            return patch_set, default_set
        except jinja2.TemplateSyntaxError as e:
            raise ConfigFormatError(
                f"The template is malformed: {e.message}")
        except KeyError:
            raise ConfigFormatError(
                "Neither the template keyword nor the 'key' keyword"
                " could not be found")


def build_scan_patches(scan_dim_patches: list, default: dict,
                       current_patch={}):
    patches = []
    if len(scan_dim_patches) > 1:
        for scan_dim_patch in scan_dim_patches[0]:
            new_cp = default
            new_cp = dtu.update_dict(current_patch, scan_dim_patch)
            patches += build_scan_patches(scan_dim_patches[1:],
                                          default,
                                          new_cp)
    else:
        for scan_dim_patch in scan_dim_patches[0]:
            patch = default
            patch = dtu.update_dict(patch, current_patch)
            patch = dtu.update_dict(patch, scan_dim_patch)
            patches.append(patch)
    return patches


def generate_patches(procedure_config):
    scan_dim_patches = []
    scan_dim_defaults = []
    if procedure_config['parameters'] is None:
        return [{}]
    for dimension in procedure_config['parameters']:
        dim_patches, dim_defaults = build_dimension_patches(dimension)
        scan_dim_patches.append(dim_patches)
        scan_dim_defaults += dim_defaults
    complete_deflt = {}
    for deflt in scan_dim_defaults:
        complete_deflt = dtu.update_dict(complete_deflt, deflt)
    return build_scan_patches(scan_dim_patches, default=complete_deflt)


def generate_init_config(procedure_config: dict):
    config = {}
    init_files = procedure_config['system_settings']['init']
    init_config_fragments = [load_configuration(icp) for icp in init_files]
    for frag in init_config_fragments:
        dtu.update_dict(config, frag, in_place=True)
    override = procedure_config['system_settings']['override']
    dtu.update_dict(config, override, in_place=True)
    return config


def generate_system_default_config(procedure_config: dict):
    config = {}
    default_config_fragments = [
        load_configuration(dcp)
        for dcp in procedure_config['system_settings']['default']]
    for frag in default_config_fragments:
        dtu.update_dict(config, frag, in_place=True)
    return config



from pathlib import Path
import logging
import sys
import queue
import threading
import click
import yaml
from rich.progress import Progress
from rich.progress import MofNCompleteColumn, SpinnerColumn
from rich.progress import TextColumn, BarColumn
from rich.progress import TimeRemainingColumn, TimeElapsedColumn
from .daq_coordination import DAQCoordClient
from .gen_configurations import generate_run_file_names


_log_level_dict = {
    "DEBUG": logging.DEBUG,
    "INFO": logging.INFO,
    "WARNING": logging.WARNING,
    "ERROR": logging.ERROR,
    "CRITICAL": logging.CRITICAL,
}


def pipelined_acquire_data(
    full_config_out: queue.Queue,
    data_acquisition_progress: queue.Queue,
    acquired_data: queue.Queue,
    full_conf_gen_done: threading.Event,
    daq_initialized: threading.Event,
    daq_done: threading.Event,
    stop: threading.Event,
    network_configuration: dict,
    initial_config: dict,
    output_dir: Path,
    keep: bool,
    readback: bool,
    start_tdcs: bool,
    full_readback: bool,
):
    # info needed for the generation of the filenames
    logger = logging.getLogger("data-acquisitor")
    logger.info("Connecting to DAQ-server")
    daq_system = DAQCoordClient(network_configuration)
    logger.info("Initializing DAQ-server")
    daq_system.initialize(initial_config, start_tdcs=start_tdcs)
    daq_initialized.set()
    logger.info("DAQ-System initialized")
    i = 0
    while (
        not full_conf_gen_done.is_set() or not full_config_out.empty()
    ) and not stop.is_set():
        run_config, file_names = full_config_out.get()
        raw_file_path = file_names[2]
        full_roc_readback_file_path = file_names[-1]
        if not keep or (keep and not raw_file_path.exists()):
            logger.info(f"gathering Data for run {i}")
            try:
                data = daq_system.measure(run_config, readback=readback)
            except ValueError as err:
                click.echo("An error ocurred during a measurement: " f"{err.args[0]}")
                del daq_system
                daq_done.set()
                return
            with open(raw_file_path, "wb+") as rdf:
                rdf.write(data)
            if full_readback:
                try:
                    config = daq_system.read_target(config=None)
                except ValueError as err:
                    click.echo(
                        "An error occured during readout of the target:"
                        f" {err.args[0]}"
                    )
                    del daq_system
                    daq_done.set()
                    return
                with open(full_roc_readback_file_path, "w+") as frcf:
                    frcf.write(yaml.safe_dump(config))
        else:
            logger.info(f"found existing data for run {i}, " "skipping acquisition")
        logger.info(f"data acquisition for run {i} done")
        logger.info(f"There are still {full_config_out.qsize()} " "items in the Queue")
        data_acquisition_progress.put(i)
        acquired_data.put(file_names)
        i += 1
    if stop.is_set():
        logger.info("Stop signal set, exiting")
    del daq_system
    daq_done.set()


@click.command()
@click.argument(
    "output_directory", type=click.Path(dir_okay=True), metavar="[OUTPUT DIR]"
)
@click.option(
    "--log",
    default="daq.log",
    type=str,
    help="Enable logging by specifying the output logfile",
)
@click.option(
    "--loglevel",
    default="INFO",
    type=click.Choice(
        ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"], case_sensitive=False
    ),
    help="specify the logging verbosity",
)
@click.option(
    "--keep/--no-keep",
    default=False,
    help="Keep the already aquired data, defaults to False",
)
@click.option(
    "--readback/--no-readback",
    default=False,
    help="Tell the sc-server to read back the register value" "written to the roc",
)
@click.option(
    "--full-readback/-no-full-readback",
    default=False,
    help="Enable a Full read of the ROC parameters after every run",
)
def pipelined_main(output_directory, log, loglevel, keep, readback, full_readback):
    logging.basicConfig(
        filename=log,
        level=_log_level_dict[loglevel],
        format="[%(asctime)s] %(levelname)s:" "%(name)-50s %(message)s",
    )
    logger = logging.getLogger("main")
    logger.info("Reading in configurations")
    # get the expected files from the directory
    output_directory = Path(output_directory)
    daq_config = output_directory / "daq_config.yaml"
    if not daq_config.exists():
        print(f"No daq_config.yaml found in {output_directory}" "exiting ...")
        sys.exit(1)
    with open(daq_config, "r", encoding='utf-8') as daqcf:
        daq_config = yaml.safe_load(daqcf.read())
    network_config = output_directory / "network_config.yaml"
    if not network_config.exists():
        print(f"No network_config.yaml found in {output_directory}" "exiting ...")
        sys.exit(1)
    with open(network_config, "r", encoding='utf-8') as nwcf:
        network_config = yaml.safe_load(nwcf.read())
    default_config = output_directory / "default_config.yaml"
    if not default_config.exists():
        print(f"No default_config.yaml found in {output_directory}" "exiting ...")
        sys.exit(1)
    with open(default_config, "r", encoding='utf-8') as dcf:
        default_config = yaml.safe_load(dcf.read())
    init_config = output_directory / "initial_state_config.yaml"
    if not init_config.exists():
        print(f"No initial_state_config.yaml found in {output_directory}" "exiting ...")
        sys.exit(1)
    with open(init_config, "r", encoding='utf-8') as icf:
        init_config = yaml.safe_load(icf.read())

    run_config_files = sorted(list(output_directory.glob("run_*_config.yaml")))
    num_digits = len(run_config_files[0].name.split("_")[1])
    run_count = len(run_config_files)
    run_ids = iter(range(len(run_config_files)))

    # read in the configurations
    # set up the different events that we are looking for
    full_gen_done = threading.Event()
    daq_system_initialized = threading.Event()
    daq_done = threading.Event()
    stop = threading.Event()

    # set up the queues to pass the data around
    full_gen_out_daq_in = queue.Queue()
    daq_progress_queue = queue.Queue()
    daq_out_unpack_in = queue.Queue()
    daq_thread = threading.Thread(
        target=pipelined_acquire_data,
        args=(
            full_gen_out_daq_in,
            daq_progress_queue,
            daq_out_unpack_in,
            full_gen_done,
            daq_system_initialized,
            daq_done,
            stop,
            network_config,
            init_config,
            output_directory,
            keep,
            readback,
            daq_config["run_start_tdc_procedure"],
            full_readback,
        ),
    )
    daq_thread.start()
    try:
        with Progress(
            SpinnerColumn(),
            TextColumn("[progress.description]{task.description}"),
            BarColumn(bar_width=None),
            MofNCompleteColumn(),
            TimeElapsedColumn(),
            TimeRemainingColumn(),
        ) as progress:
            read_configurations = progress.add_task(
                "[cyan] Read run configurations from disk", total=run_count
            )
            daq_progbar = progress.add_task(
                "[turquoise] Acquiring Data from the Test System",
                total=run_count,
                start=False,
            )
            daq_active = False
            while not daq_done.is_set():
                # activate the progress bar when daq is initialized
                if daq_system_initialized.is_set() and not daq_active:
                    progress.start_task(daq_progbar)
                    daq_active = True
                try:
                    run_id = next(run_ids)
                    run_files = generate_run_file_names(
                        run_id, num_digits, output_directory
                    )
                    with open(run_files[0], "r", encoding="utf-8") as patchfile:
                        patch = yaml.safe_load(patchfile.read())
                    full_gen_out_daq_in.put((patch, run_files))
                    progress.update(read_configurations, advance=1)
                except StopIteration:
                    full_gen_done.set()
                    pass
                try:
                    _ = daq_progress_queue.get(block=False)
                    progress.update(daq_progbar, advance=1)
                except queue.Empty:
                    pass
                try:
                    _ = daq_out_unpack_in.get(block=False)
                except queue.Empty:
                    pass
    except KeyboardInterrupt:
        stop.set()
    daq_thread.join()



from copy import deepcopy
from .config_errors import ConfigPatchError


def patch_configuration(config: dict, patch: dict):
    """
    currently only a wrapper function, if tests need to be implemented,
    then this is done here. Also the correct part of the cofiguration
    may be selected
    """
    return update_dict(config, patch)


def update_dict(original: dict, update: dict, offset: bool = False,
                in_place: bool = False):
    """
    update one multi level dictionary with another. If a key does not
    exist in the original, it is created. the updated dictionary is returned

    Arguments:
    original: the unaltered dictionary

    update: the dictionary that will either overwrite or extend the original

    offset: If True, the values in the update dictionary will be interpreted as
            offsets to the values in the original dict. The resulting value wil
            be original + update (this will be string concatenation if both the
            original and update values are strings

    Returns:
    updated_dict: the extended/updated dict where the values that do not appear
                  in the update dict are from the original dict, the keys that
                  do appear in the update dict overwrite/extend the values in
                  the original
    Raises:
    TypeError: If offset is set to true and the types don't match the addition
               of the values will cause a TypeError to be raised

    ConfigPatchError: If lists are updated the length of the lists needs to
               match otherwise a error is raised
    """

    if in_place is True:
        result = original
    else:
        result = deepcopy(original)
    for update_key in update.keys():
        if update_key not in result.keys():
            if in_place is True:
                result[update_key] = update[update_key]
            else:
                result[update_key] = deepcopy(update[update_key])
            continue
        if not isinstance(result[update_key], type(update[update_key])):

            if in_place is True:
                result[update_key] = update[update_key]
            else:
                result[update_key] = deepcopy(update[update_key])
            continue
        if isinstance(result[update_key], dict):
            result[update_key] = update_dict(result[update_key],
                                             update[update_key],
                                             offset=offset,
                                             in_place=in_place)
        elif isinstance(result[update_key], list) or\
                isinstance(result[update_key], tuple):
            if len(result[update_key]) != len(update[update_key]):
                raise ConfigPatchError(
                    'If a list is a value of the dict the list'
                    ' lengths in the original and update need to'
                    ' match')
            patch = []
            for i, (orig_elem, update_elem) in enumerate(
                    zip(result[update_key], update[update_key])):
                if isinstance(orig_elem, dict):
                    patch.append(update_dict(result[update_key][i],
                                             update_elem,
                                             offset=offset,
                                             in_place=in_place))
                else:
                    if in_place is True:
                        patch.append(update_elem)
                    else:
                        patch.append(deepcopy(update_elem))
            if isinstance(result[update_key], tuple):
                patch = tuple(patch)
            result[update_key] = patch
        else:
            if offset is True:
                result[update_key] += update[update_key]
            else:
                if in_place is True:
                    result[update_key] = update[update_key]
                else:
                    result[update_key] = deepcopy(update[update_key])
    return result


def diff_dict(d1: dict, d2: dict) -> dict:
    """
    create the diff of two dictionaries.

    Create a dictionary containing all the keys that differ between
    d1 and d2 or that exist in d2 but not in d1. The value of the
    differing key will be the value found in d2.

    Example:
        With the inputs being
        d1 = {'a': 1, 'b': {'c': 2, 'f': 4}, 'e': 3}
        d2 = {'a': 2, 'b': {'c': 3, 'f': 4}, 'e': 3, 'g': 4}

        the resulting diff will be
        diff = {'a': 2, 'b': {'c':3}, 'g': 4}

        showing that the keys 'a' and 'b:c' are different and that the
        value in the differing d2 is 2 for a and 3 for 'b:c' respectively.
        As there is an entry in d2 that is not in d1 it will appear in the
        diff, however if there is an entry in d1 that is not in d2 it will
        not appear.

    This function is intended to find the patch that where applied on d1 to
    create d2. This may provide some motivation for the way that the function
    handles the different dictionaries
    """
    diff = {}
    for key2, value2 in d2.items():
        if key2 not in d1.keys():
            diff[key2] = value2
        elif isinstance(d1[key2], dict) and isinstance(value2, dict):
            potential_diff = diff_dict(d1[key2], value2)
            if potential_diff is not None:
                diff[key2] = potential_diff
        elif value2 != d1[key2]:
            diff[key2] = value2
    if diff == {}:
        return None
    return diff



"""
Utilities for the use with the handling of the gathered data
in the datenraffinerie.
"""
import subprocess as sp
import numpy as np
from pathlib import Path
import os
import shutil
import pandas as pd
import tables
import yaml
import logging


class AnalysisError(Exception):
    def __init__(self, message):
        super().__init__(message)


def cartesian(arrays, dtype=np.float32):
    n = 1
    for i in range(len(arrays)):
        arrays[i] = np.array(arrays[i], dtype=dtype)
    for x in arrays:
        n *= x.size
    out = np.zeros((n, len(arrays)), dtype=dtype)

    for i in range(len(arrays)):
        m = int(n / arrays[i].size)
        out[:n, i] = np.repeat(arrays[i], m)
        n //= arrays[i].size

    n = arrays[-1].size
    for k in range(len(arrays)-2, -1, -1):
        n *= arrays[k].size
        m = int(n / arrays[k].size)
        for j in range(1, arrays[k].size):
            out[j*m:(j+1)*m, k+1:] = out[0:m, k+1:]
    return out


def read_whole_dataframe(pytables_filepath):
    dfile = tables.open_file(pytables_filepath, 'r')
    table = dfile.root.data.measurements
    df = pd.DataFrame.from_records(table.read())
    dfile.close()
    return df


def read_dataframe_chunked(pytables_filepaths):
    for filepath in pytables_filepaths:
        dfile = tables.open_file(filepath)
        table = dfile.root.data.measurements
        yield pd.DataFrame.from_records(table.read())
        dfile.close()


def start_compiled_fracker(rootfile: Path,
                           hdf_file: Path,
                           full_config_path: Path,
                           raw_data: bool,
                           columns: list,
                           compression: int,
                           logger: logging.Logger,
                           block_size: int = 2000000):
    """
    Assuming there is a fracker command that merges in the configuration
    and transforms the root into the hdf file call it with the proper arguments
    """
    compiled_fracker_path = shutil.which('fracker')
    if compiled_fracker_path is None:
        raise FileNotFoundError('The compiled fracker could not be found')
    # build the column list if none has been provided
    if columns is None:
        if raw_data:
            columns = event_mode_data_columns.copy()
        else:
            columns = data_columns.copy()
        columns += expected_columns

    # assemble the fracker command string passed to the system
    fracker_command_tokens = [compiled_fracker_path]
    fracker_command_tokens.append('-c')
    fracker_command_tokens.append(str(full_config_path.absolute()))
    fracker_command_tokens.append('-i')
    fracker_command_tokens.append(str(rootfile.absolute()))
    fracker_command_tokens.append('-o')
    fracker_command_tokens.append(str(hdf_file.absolute()))
    fracker_command_tokens.append('-b')
    fracker_command_tokens.append(str(block_size))
    fracker_command_tokens.append('-s')
    fracker_command_tokens += [str(col) for col in columns]
    fracker_command_tokens.append('-p')
    fracker_command_tokens.append(str(compression))
    fracker_command_tokens.append('-f')
    fracker_command_tokens.append('full' if raw_data else 'summary')
    logger.info(
        'Ran fracker with command: ' + ' '.join(fracker_command_tokens))
    return sp.Popen(fracker_command_tokens, stdout=sp.PIPE)


def run_turbo_pump(output_file: str, input_files: list):
    """
    run the compiled turbo pump to merge files
    """
    turbo_pump_path = shutil.which('turbo-pump')
    tp_output = " -o " + str(output_file)
    tp_input = " -i"
    for inp in input_files:
        tp_input += " " + str(inp)
    full_turbo_p_command = turbo_pump_path + tp_output + tp_input
    return os.system(full_turbo_p_command)


def start_unpack(raw_path,
                 unpacked_path,
                 logger: logging.Logger,
                 raw_data: bool = False,
                 characMode: bool = False):
    # we need to create a meta-yaml file for the unpacker
    meta_yaml_path = os.path.splitext(
            os.path.abspath(unpacked_path))[0] + '_unpack_meta.yaml'
    meta_yaml = {}
    meta_yaml['metaData'] = {}
    meta_yaml['metaData']['characMode'] = 1 if characMode else 0
    meta_yaml['metaData']['keepRawData'] = 1 if raw_data else 0
    meta_yaml['metaData']['keepSummary'] = 0 if raw_data else 1
    meta_yaml['metaData']['chip_params'] = {}
    with open(meta_yaml_path, 'w+') as mcf:
        mcf.write(yaml.safe_dump(meta_yaml))

    unpack_command_path = shutil.which('unpack')
    if unpack_command_path is None:
        raise FileNotFoundError('Unable to find the unpack command')
    unpack_command_tokens = ['unpack']
    # add the input file to the token list
    unpack_command_tokens.append('-i')
    unpack_command_tokens.append(str(raw_path))
    # add the output file to the token list
    unpack_command_tokens.append('-o')
    unpack_command_tokens.append(str(unpacked_path))
    unpack_command_tokens.append('-M')
    unpack_command_tokens.append(str(meta_yaml_path))
    logger.info('starting unpack command: ' + ' '.join(unpack_command_tokens))
    return sp.Popen(unpack_command_tokens, stdout=sp.PIPE)


def compute_channel_type_from_event_data(chip: int, channel: int, half: int):
    if channel <= 35:
        out_channel = channel * (half + 1)
        out_type = 0
    if channel == 36:
        out_channel = half
        out_type = 1
    if channel > 36:
        out_channel = channel - 37 + (half * 2)
        out_type = 100
    return chip, out_channel, out_type


event_mode_data_columns = [
        'event', 'chip', 'half', 'channel', 'adc', 'adcm', 'toa',
        'tot', 'totflag', 'trigtime', 'trigwidth', 'corruption',
        'bxcounter', 'eventcounter', 'orbitcounter']

data_columns = ['chip', 'channel', 'channeltype',
                'adc_median', 'adc_iqr', 'tot_median',
                'tot_iqr', 'toa_median', 'toa_iqr', 'adc_mean',
                'adc_stdd', 'tot_mean',
                'tot_stdd', 'toa_mean', 'toa_stdd', 'tot_efficiency',
                'tot_efficiency_error', 'toa_efficiency',
                'toa_efficiency_error']
expected_columns = [
    'Adc_pedestal', 'Channel_off', 'DAC_CAL_CTDC_TOA', 'DAC_CAL_CTDC_TOT',
    'DAC_CAL_FTDC_TOA', 'DAC_CAL_FTDC_TOT', 'DIS_TDC', 'ExtData',
    'HZ_inv', 'HZ_noinv', 'HighRange', 'IN_FTDC_ENCODER_TOA',
    'IN_FTDC_ENCODER_TOT', 'Inputdac', 'LowRange', 'mask_AlignBuffer',
    'mask_adc', 'mask_toa', 'mask_tot', 'probe_inv',
    'probe_noinv', 'probe_pa', 'probe_toa', 'probe_tot',
    'sel_trig_toa', 'sel_trig_tot', 'trim_inv', 'trim_toa',
    'trim_tot', 'Adc_TH', 'Bx_offset', 'CalibrationSC',
    'ClrAdcTot_trig', 'IdleFrame', 'L1Offset', 'MultFactor',
    'SC_testRAM', 'SelTC4', 'Tot_P0', 'Tot_P1',
    'Tot_P2', 'Tot ', 'Tot_P_Add',
    'Tot_TH0', 'Tot_TH1', 'Tot_TH2', 'Tot_TH3',
    'sc_testRAM', 'Cf', 'Cf_comp', 'Clr_ADC',
    'Clr_ShaperTail', 'Delay40', 'Delay65', 'Delay87',
    'Delay9', 'En_hyst_tot', 'Ibi_inv', 'Ibi_inv_buf',
    'Ibi_noinv', 'Ibi_noinv_buf', 'Ibi_sk', 'Ibo_inv',
    'Ibo_inv_buf', 'Ibo_no ', 'Ibo_noinv_buf',
    'Ibo_sk', 'ON ', 'ON_ref_adc',
    'ON_rtr', 'ON_toa', 'ON_tot', 'Rc',
    'Rf', 'S_inv', 'S_inv_buf',
    'S_noinv', 'S_noinv_buf', 'S_sk', 'SelExtADC',
    'SelRisingEdge', 'dac_pol', 'gain_tot', 'neg',
    'pol_trig_toa', 'range_indac', 'range_inv', 'range_tot',
    'ref_adc', 'trim_vbi_pa', 'trim_vbo_pa', 'BIAS_CAL_DAC_CTDC_P_D',
    'BIAS_CAL_DAC_CTDC_P_EN', 'BIAS_FOLLOWER_CAL_P_CTDC_EN',
    'BIAS_FOLLOWER_CAL_P_D', 'BIAS_FOLLOWER_CAL_P_FTDC_D',
    'BIAS_FOLLOWER_CAL_P_FTDC_EN', 'BIAS_I_CTDC_D',
    'BIAS_I_FTDC_D', 'CALIB_CHANNEL_DLL',
    'CTDC_CALIB_FREQUENCY', 'CTRL_IN_REF_CTDC_P_D',
    'CTRL_IN_REF_CTDC_P_EN', 'CTRL_IN_REF_FTDC_P_D',
    'CTRL_IN_REF_FTDC_P_EN', 'CTRL_IN_SIG_CTDC_P_D',
    'CTRL_IN_SIG_CTDC_P_EN', 'CTRL_IN_SIG_FTDC_P_D',
    'CTRL_IN_SIG_FTDC_P_EN', 'EN_MASTER_CTDC_DLL',
    'EN_MASTER_CTDC_VOUT_INIT', 'EN_MASTER_FTDC_DLL',
    'EN_MASTER_FTDC_VOUT_INIT', 'EN_REF_BG',
    'FOLLOWER_CTDC_EN', 'FOLLOWER_FTDC_EN',
    'FTDC_CALIB_FREQUENCY', 'GLOBAL_DISABLE_TOT_LIMIT',
    'GLOBAL_EN_BUFFER_CTDC', 'GLOBAL_EN_BUFFER_FTDC',
    'GLOBAL_EN_TOT_PRIORITY', 'GLOBAL_EN_TUNE_GAIN_DAC',
    'GLOBAL_FORCE_EN_CLK', 'GLOBAL_FORCE_EN_OUTPUT_DATA',
    'GLOBAL_FORCE_EN_TOT', 'GLOBAL_INIT_DAC_B_CTDC',
    'GLOBAL_LATENCY_TIME', 'GLOBAL_MODE_FTDC_TOA',
    'GLOBAL_MODE_NO_TOT_SUB', 'GLOBAL_MODE_TOA_DIRECT_OUTPUT',
    'GLOBAL_SEU_TIME_OUT', 'GLOBAL_TA_SELECT_GAIN_TOA',
    'GLOBAL_TA_SELECT_GAIN_TOT', 'INV_FRONT_40MHZ',
    'START_COUNTER', 'VD_CTDC_N_D',
    'VD_CTDC_N_DAC_EN', 'VD_CTDC_N_FORCE_MAX',
    'VD_CTDC_P_D', 'VD_CTDC_P_DAC_EN',
    'VD_FTDC_N_D', 'VD_FTDC_N_DAC_EN',
    'VD_FTDC_N_FORCE_MAX', 'VD_FTDC_P_D',
    'VD_FTDC_P_DAC_EN', 'sel_clk_rcg', 'Calib', 'ExtCtest',
    'IntCtest', 'Inv_vref', 'Noinv_vref', 'ON_dac',
    'Refi', 'Toa_vref', 'Tot_vref', 'Vbg_1v',
    'probe_dc', 'probe_dc1', 'probe_dc2', 'BIAS_I_PLL_D',
    'DIV_PLL', 'EN', 'EN_HIGH_CAPA', 'EN_LOCK_CONTROL',
    'EN_PLL', 'EN_PhaseShift', 'EN_RCG', 'EN_REF_BG',
    'EN_probe_pll', 'ENpE', 'ERROR_LIMIT_SC', 'EdgeSel_T1',
    'FOLLOWER_PLL_EN', 'INIT_D', 'INIT_DAC_EN', 'Pll_Locked_sc',
    'PreL1AOffset', 'RunL', 'RunR', 'S',
    'TestMode', 'VOUT_INIT_EN', 'VOUT_INIT_EXT_D', 'VOUT_INIT_EXT_EN',
    'b_in', 'b_out', 'err_countL', 'err_countR',
    'fc_error_count', 'in_inv_cmd_rx', 'lock_count', 'n_counter_rst',
    'phase_ck', 'phase_strobe', 'rcg_gain', 'sel_40M_ext',
    'sel_error', 'sel_lock', 'sel_strobe_ext', 'srout',
    'statusL', 'statusR', 'Tot_P3', 'Ibo_noinv', 'ON_pa'
    ]

daq_columns = ['Bx_trigger',
               'A_enable',
               'B_enable',
               'C_enable',
               'D_enable',
               'A_BX',
               'B_BX',
               'C_BX',
               'D_BX',
               'A_length',
               'B_length',
               'C_length',
               'D_length',
               'A_prescale',
               'B_prescale',
               'C_prescale',
               'D_prescale',
               ]



from setuptools import setup

setup(
    entry_points={
        'console_scripts': [
            'datenraffinerie = datenraffinerie.datenraffinerie:cli',
            'generate-configs = datenraffinerie.gen_configurations:generate_configuratons',
            'acquire-data-single-threaded = datenraffinerie.acquire_data:acquire_data',
            'acquire-data = datenraffinerie.acquire_data:pipelined_main',
            'coordinate-daq-access = datenraffinerie.daq_coordination:main',
            'process-raw-data-single-threaded = datenraffinerie.frack_data:main',
            'process-raw-data = datenraffinerie.postprocessing_queue:main',
            'yaml-utils = datenraffinerie.yaml_utils:cli',
            'show-hdf = datenraffinerie.print_h5:show_hdf',
            'read-rocs = datenraffinerie.read_rocs:cli',
            'full-daq = datenraffinerie.full_daq:main',
        ]
    },
    install_requires=[
        'luigi',
        'pandas',
        'matplotlib',
        'numpy',
        'scipy',
        'uproot',
        'pyyaml',
        'zmq',
        'pytest',
        'awkward',
        'tables',
        'h5py',
        'numba',
        'jinja2',
        'schema',
        'hgcroc-configuration-client',
        'progress',
        'pymongo',
        'rich',
    ]
)



import os
import sys
import math
import shutil
from pathlib import Path
import threading
import multiprocessing as mp
import queue
import logging
from copy import deepcopy
from typing import Iterator, Tuple, Generator
import click
import yaml
from rich.progress import Progress
from rich.progress import SpinnerColumn
from rich.progress import MofNCompleteColumn
from rich.progress import TextColumn
from rich.progress import BarColumn
from rich.progress import TimeElapsedColumn, TimeRemainingColumn
from . import config_utilities as cfu
from . import dict_utils as dctu

_log_level_dict = {
    "DEBUG": logging.DEBUG,
    "INFO": logging.INFO,
    "WARNING": logging.WARNING,
    "ERROR": logging.ERROR,
    "CRITICAL": logging.CRITICAL,
}


def generate_tool_configurations(
    output_dir: Path,
    network_config: Path,
    procedure: dict,
    system_default_config: dict,
    system_init_config: dict,
) -> None:
    """
    Create the configurations that govern the behaviour of the tools along with
    the initial state configurations of the system
    """
    init_config_path = output_dir / "initial_state_config.yaml"
    default_config_path = output_dir / "default_config.yaml"
    post_config_path = output_dir / "postprocessing_config.yaml"
    daq_config_path = output_dir / "daq_config.yaml"
    proc_network_config = output_dir / "network_config.yaml"
    procedure_config_files = [
        init_config_path,
        default_config_path,
        post_config_path,
        daq_config_path,
    ]
    # clean the output directory of any previous configuration files
    for file in procedure_config_files:
        if file.exists():
            os.remove(file)
    # generate the initial, default and network configuration
    shutil.copyfile(network_config, proc_network_config)
    with open(default_config_path, "w+", encoding="utf-8") as dcf:
        dcf.write(yaml.safe_dump(system_default_config))
    with open(init_config_path, "w+", encoding="utf-8") as icf:
        icf.write(yaml.safe_dump(system_init_config))
    with open(post_config_path, "w+", encoding="utf-8") as pcf:
        post_config = {}
        post_config["data_columns"] = procedure["data_columns"]
        post_config["mode"] = procedure["mode"]
        post_config["procedure"] = procedure["name"]
        post_config["data_format"] = procedure["data_format"]
        pcf.write(yaml.safe_dump(post_config))
    with open(daq_config_path, "w+", encoding="utf-8") as daqcf:
        daq_config = {}
        daq_config["run_start_tdc_procedure"] = procedure["run_start_tdc_procedure"]
        daqcf.write(yaml.safe_dump(daq_config))
    return


def generate_run_file_names(
    run_id: int, width: int, output_dir: Path
) -> Tuple[Path, Path, Path, Path, Path, Path]:
    """
    Generate the file names for all the files of the run
    """
    run_conf_file_name = f"run_{run_id:0>{width}}_config.yaml"
    full_run_conf_file_name = f"run_{run_id:0>{width}}_config_full.yaml"
    readback_run_conf_file_name = f"run_{run_id:0>{width}}_config_readback.yaml"
    raw_file_name = f"run_{run_id:0>{width}}_data.raw"
    root_file_name = f"run_{run_id:0>{width}}_data.root"
    hdf_file_name = f"run_{run_id:0>{width}}_data.h5"
    run_conf_path = output_dir / run_conf_file_name
    raw_file_path = output_dir / raw_file_name
    root_file_path = output_dir / root_file_name
    hdf_file_path = output_dir / hdf_file_name
    full_run_conf_path = output_dir / full_run_conf_file_name
    readback_config_path = output_dir / readback_run_conf_file_name
    return (
        run_conf_path,
        full_run_conf_path,
        raw_file_path,
        root_file_path,
        hdf_file_path,
        readback_config_path,
    )


def write_patch_to_file(
    config: dict, filenames: Tuple[Path, Path, Path, Path, Path, Path]
) -> Tuple[dict, Tuple[Path, Path, Path, Path, Path, Path]]:
    """
    Write the patch to a file and return what was
    passed in to act as a map function with side effects
    """
    with open(filenames[0], "w+", encoding="utf-8") as patch_file:
        patch_file.write(yaml.safe_dump(config))
    return (config, filenames)


def pipelined_generate_patches(
    output_dir: Path,
    run_configs: Iterator,
    patch_out_ful_config_in_queue: queue.Queue,
    config_gen_progress_queue: queue.Queue,
    patch_gen_done: threading.Event,
    num_digits: int,
    stop: threading.Event,
):
    """
    Generate the files in the output_dir that contain the patches
    that are applied to the system and pass the patch + the run file names down
    the queue
    """
    patch_generation_pipeline = map(
        lambda x: write_patch_to_file(*x),
        map(
            lambda x: (x[1], generate_run_file_names(x[0], num_digits, output_dir)),
            enumerate(run_configs),
        ),
    )
    for filepath_set in patch_generation_pipeline:
        if stop.is_set():
            break
        patch_out_ful_config_in_queue.put(filepath_set)
        config_gen_progress_queue.put(1)
    patch_gen_done.set()


def write_out_full_config(
    patch: dict,
    full_config: dict,
    run_file_paths: Tuple[Path, Path, Path, Path, Path, Path],
    config_out_queue: mp.Queue,
    config_progress_queue: mp.Queue,
) -> None:
    """
    Write the output of the full config to a file and write signal
    to the queues that this has been done
    """
    with open(run_file_paths[1], "w+", encoding="utf-8") as out_file:
        out_file.write(yaml.safe_dump(full_config))
    config_out_queue.put((patch, run_file_paths))
    config_progress_queue.put(1)


def full_config_param_generator(
    patch_gen_done: threading.Event,
    full_config: dict,
    patch_gen_output: queue.Queue,
    counfig_out_queue: mp.Queue,
    config_gen_progress_queue: mp.Queue,
    stop: threading.Event,
) -> Generator:
    """
    transform the queue of the patch write thread into a generator and
    keep track of the state of the config that is to be written out
    """
    current_config = deepcopy(full_config)
    while not patch_gen_done.is_set() or not patch_gen_output.empty():
        config, run_files = patch_gen_output.get()
        current_config = dctu.update_dict(current_config, config, in_place=True)
        yield (
            config,
            current_config,
            run_files,
            counfig_out_queue,
            config_gen_progress_queue,
        )
        if stop.is_set():
            break


def pipelined_generate_full_run_config(
    patch_gen_done: threading.Event,
    patch_gen_out_queue: queue.Queue,
    full_config: dict,
    full_conf_gen_out_daq_in_queue: mp.Queue,
    config_gen_progress_queue: mp.Queue,
    full_conf_gen_done: threading.Event,
    num_generators: int,
    stop: threading.Event,
):
    """
    Produce the full config and write it onto disk in parallel to
    speed up the process
    """
    with mp.Pool(num_generators) as config_gen_pool:
        config_gen_pool.starmap(
            write_out_full_config,
            full_config_param_generator(
                patch_gen_done,
                full_config,
                patch_gen_out_queue,
                full_conf_gen_out_daq_in_queue,
                config_gen_progress_queue,
                stop,
            ),
        )
    full_conf_gen_done.set()


@click.command()
@click.argument(
    "config", type=click.Path(exists=True), metavar="[main configuration file]"
)
@click.argument("netcfg", type=click.Path(exists=True))
@click.argument(
    "procedure", type=str, metavar="[Procedure to be run by the datenraffinerie]"
)
@click.argument(
    "output_dir",
    type=click.Path(dir_okay=True),
    metavar="[Location to write the configuration files to]",
)
@click.option(
    "--full_conf_generators",
    "-f",
    type=int,
    default=1,
    help="set how many full-cfg generators to run, default 1",
)
@click.option("--log/--no-log", type=bool, default=True, help="Enable/Disable logging")
@click.option(
    "--loglevel",
    default="INFO",
    type=click.Choice(
        ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"], case_sensitive=False
    ),
    help="specify the logging verbosity",
)
def generate_configuratons(
    config, netcfg, procedure, output_dir, full_conf_generators, log, loglevel
):
    # generate the conifgurations
    if log:
        logging.basicConfig(
            filename="gen_config.log",
            level=loglevel,
            format="[%(asctime)s] %(levelname)s:" "%(name)-50s %(message)s",
        )
    config = click.format_filename(config)
    try:
        procedure, (
            system_default_config,
            system_init_config,
            run_configs,
            run_count,
        ) = cfu.get_procedure_configs(
            main_config_file=config,
            procedure_name=procedure,
            calibration=None,
            diff=True,
        )
    except ValueError as err:
        print(f"The procedure with name: {err.args[1]} could not be found,")
        print("Available procedures are:")
        for pname in err.args[2]:
            print(f"\t {pname}")
        sys.exit(1)

    # create the output directory and the initial files
    output_dir = Path(output_dir)
    if not os.path.exists(output_dir):
        os.mkdir(output_dir)

    generate_tool_configurations(
        output_dir, netcfg, procedure, system_default_config, system_init_config
    )

    # generate the configurations for the runs
    num_digits = math.ceil(math.log(run_count, 10))
    full_config = dctu.update_dict(system_default_config, system_init_config)

    # queues that hold the ouptut of the patch generator thread
    patch_gen_out_full_gen_in = queue.Queue()
    patch_gen_progress_queue = queue.Queue()
    patch_gen_done = threading.Event()

    # generate the starting config for multiple full_conf generators
    fcm = mp.Manager()
    full_configs_generated = fcm.Queue()
    full_config_generation_progress = fcm.Queue()
    full_config_generation_done = threading.Event()

    # global signals
    stop = mp.Event()

    config_gen_thread = threading.Thread(
        target=pipelined_generate_patches,
        args=(
            output_dir,
            run_configs,
            patch_gen_out_full_gen_in,
            patch_gen_progress_queue,
            patch_gen_done,
            num_digits,
            stop,
        ),
    )
    full_config_gen_thread = threading.Thread(
        target=pipelined_generate_full_run_config,
        args=(
            patch_gen_done,
            patch_gen_out_full_gen_in,
            full_config,
            full_configs_generated,
            full_config_generation_progress,
            full_config_generation_done,
            full_conf_generators,
            stop,
        ),
    )
    config_gen_thread.start()
    full_config_gen_thread.start()
    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        BarColumn(bar_width=None),
        MofNCompleteColumn(),
        TimeElapsedColumn(),
        TimeRemainingColumn(),
    ) as progress:
        run_progress_bar = progress.add_task(
            "Generating run configurations", total=run_count
        )
        full_progress_bar = progress.add_task(
            "Generating full config for fracker", total=run_count
        )
        while (
            not full_config_generation_done.is_set()
            or not full_config_generation_progress.empty()
        ):
            while True:
                try:
                    _ = patch_gen_progress_queue.get(block=False)
                    progress.update(run_progress_bar, advance=1)
                except queue.Empty:
                    break
            while True:
                try:
                    _ = full_configs_generated.get(block=False)
                except queue.Empty:
                    break
            while True:
                try:
                    _ = full_config_generation_progress.get(block=False)
                    progress.update(full_progress_bar, advance=1)
                except queue.Empty:
                    break
            logging.debug(
                f"{full_config_generation_done.is_set()}"
                " = state of full config generator"
            )
        progress.refresh()
    config_gen_thread.join()
    full_config_gen_thread.join()



class OutputError(Exception):
    def __init__(self, message):
        super().__init__(message)
        self.message = message

class DAQError(Exception):
    def __init__(self, message):
        self.message = message

class DAQConfigError(Exception):
    def __init__(self, message):
        self.message = message





import click
import yaml
from rich import print
from hgcroc_configuration_client.client import Client as SCClient
import sys


@click.group("cli")
@click.pass_context
def cli(ctx):
    if ctx.invoked_subcommand is None:
        click.echo('No command was given to read-rocs. Available commands are:'
                   '\n\tconfigure\n\tread\n\treset\n\tdescribe')


@cli.command()
@click.argument('hostname', type=str,
                metavar='[HOSTNAME]')
@click.argument('port', type=str,
                metavar='[PORT]')
@click.argument('configuration', type=click.File('r'))
@click.option(
        '--readback',
        '-r',
        is_flag=True,
        show_default=True,
        default=False,
        help='read back every parameter written to the '
             'ROC to make sure that the config was written properly')
def configure(hostname, port, configuration, readback):
    sc_client = SCClient(hostname, port=port)
    try:
        config = yaml.safe_load(configuration.read())
    except yaml.ScannerError as err:
        print(f'Could not read in yaml. {err}')
        sys.exit(1)
    try:
        config = config['target']
    except KeyError:
        pass
    try:
        sc_client.set(config, readback=readback)
    except ValueError as err:
        print('An error occured during the writing of '
              f'the configuration: {err}')
        sys.exit(2)
    print('Configuration written to the rocs')


@cli.command()
@click.argument('hostname', type=str,
                metavar='[HOSTNAME]')
@click.argument('port', type=str,
                metavar='[PORT]')
@click.option('--output', '-o', type=click.Path(), default=None,
              help='File that the Config is written is written to. '
                   'If no file is given the output is printed on stdout')
def read(hostname, port, output):
    sc_client = SCClient(hostname, port=port)
    try:
        config = sc_client.get_from_hardware(config=None)
    except ValueError as err:
        print(f'During reading of the ROCs an error occured: {err}')
        sys.exit(3)
    printout = yaml.safe_dump({'target': config})
    if output is None:
        print(printout)
    else:
        with open(output, 'w+') as rcof:
            rcof.write(printout)


@cli.command()
@click.argument('hostname', type=str,
                metavar='[HOSTNAME]')
@click.argument('port', type=str,
                metavar='[PORT]')
def reset(hostname, port):
    sc_client = SCClient(hostname, port=port)
    try:
        sc_client.reset()
    except ValueError as err:
        print(f'During reset an error occured: {err}')
        sys.exit(4)


@cli.command()
@click.argument('hostname', type=str,
                metavar='[HOSTNAME]')
@click.argument('port', type=str,
                metavar='[PORT]')
@click.option('--output', '-o', type=click.Path(), default=None,
              help='File that the description is written is written to. '
                   'If no file is given the output is printed on stdout')
def describe(hostname, port, output):
    sc_client = SCClient(hostname, port=port)
    try:
        description = sc_client.describe()
    except ValueError as err:
        print(f'During the readout of the description an error occured: {err}')
        sys.exit(5)
    printout = yaml.safe_dump({'target': description})
    if output is None:
        print(printout)
    else:
        with open(output, 'w+') as descf:
            descf.write(printout)



from . import analysis_utilities as anu
from . import dict_utils as dcu
from progress.bar import Bar
from time import sleep
import glob
import os
import logging
from pathlib import Path
import yaml
import click
import sys


def unpack_in_parallel(waiting_tasks, parallel_tasks,
                       logger, bar: bool, mode: str):
    running_tasks = []
    failed_tasks = []
    successful_tasks = []
    if bar:
        prog_bar = Bar('creating intermediary root files'.ljust(50, ' '),
                       max=len(waiting_tasks))
    while len(waiting_tasks) > 0 or len(running_tasks) > 0:
        # check whether there another task can be run
        while parallel_tasks > len(running_tasks):
            try:
                (raw_path, unpacked_path, formatted_path, full_config_path) = \
                        waiting_tasks.pop()
            except IndexError:
                break
            if mode == 'full':
                raw_data = True
            else:
                raw_data = False
            running_tasks.append((
                anu.start_unpack(raw_path,
                                 unpacked_path,
                                 logging.getLogger(f'unpack-{raw_path}'),
                                 raw_data=raw_data),
                raw_path,
                unpacked_path,
                formatted_path,
                full_config_path
                )
            )
        del_indices = []
        for i, (task, raw_path, unpacked_path,
                formatted_path, full_config_path) \
                in enumerate(running_tasks):
            returncode = task.poll()
            if returncode is None:
                continue
            del_indices.append(i)
            if returncode == 0:
                logger.info(f'created root file from {raw_path}')
                successful_tasks.append(
                        (unpacked_path, formatted_path, full_config_path)
                )
            if returncode > 0:
                logger.error(f'failed to created root file from {raw_path}, '
                             f'unpacker returned error code {returncode}')
                failed_tasks.append((raw_path, unpacked_path))
            if returncode < 0:
                logger.error(f'failed to created root file from {raw_path}, '
                             'unpacker got terminated')
                failed_tasks.append((raw_path, unpacked_path))
            if bar:
                prog_bar.next()
        del_indices = sorted(del_indices, reverse=True)
        for i in del_indices:
            del running_tasks[i]
        sleep(.2)
    if bar:
        prog_bar.finish()
    return successful_tasks, failed_tasks


def frack_in_parallel(waiting_tasks, parallel_tasks,
                      logger, bar, mode: str, columns,
                      compression):
    running_tasks = []
    failed_tasks = []
    successful_tasks = []
    if bar:
        prog_bar = Bar('creating hdf files'.ljust(50, ' '),
                       max=len(waiting_tasks))
    while len(waiting_tasks) > 0 or len(running_tasks) > 0:
        # check whether there another task can be run
        while parallel_tasks > len(running_tasks):
            try:
                (unpacked_path, formatted_path, full_config_path) = \
                        waiting_tasks.pop()
            except IndexError:
                break
            if mode == 'full':
                raw_data = True
            else:
                raw_data = False
            running_tasks.append((
                anu.start_compiled_fracker(unpacked_path, formatted_path,
                                           full_config_path, raw_data,
                                           columns, compression, logger),
                unpacked_path,
                formatted_path,
                full_config_path
                )
            )
        del_indices = []
        for i, (task, unpacked_path, formatted_path, full_config_path) \
                in enumerate(running_tasks):
            returncode = task.poll()
            if returncode is None:
                continue
            del_indices.append(i)
            if returncode == 0:
                logger.info(
                        f'created {formatted_path} file from {unpacked_path}')
                successful_tasks.append((unpacked_path, formatted_path))
            if returncode > 0:
                logger.error(
                        f'failed to created hdf file from {formatted_path}, '
                        f'fracker returned error code {returncode}')
                failed_tasks.append((unpacked_path, formatted_path))
            if returncode < 0:
                logger.error(
                        f'failed to created hdf file from {formatted_path}, '
                        'fracker got terminated')
                failed_tasks.append((unpacked_path, formatted_path))
            if bar:
                prog_bar.next()
        del_indices = sorted(del_indices, reverse=True)
        for i in del_indices:
            del running_tasks[i]
        sleep(.2)
    if bar:
        prog_bar.finish()
    return successful_tasks, failed_tasks


def frack_data(raw_data_paths, full_run_config_paths, mode, columns,
               parallel_tasks: int = 2,
               bar: bool = False, compression: bool = 3):
    logger = logging.getLogger('fracker')
    # generate the paths for the output files
    unpacked_data_paths = [Path(os.path.splitext(rdfp)[0] + '.root')
                           for rdfp in raw_data_paths]
    formatted_data_paths = [Path(os.path.splitext(rdfp)[0] + '.h5')
                            for rdfp in raw_data_paths]

    # generate the intermediary root files
    logger.info('Converting raw files into root files')
    unpack_tasks = list(
            zip(raw_data_paths, unpacked_data_paths,
                formatted_data_paths, full_run_config_paths)
    )
    unpack_tasks.reverse()
    try:
        successful_tasks, failed_tasks = \
            unpack_in_parallel(unpack_tasks, parallel_tasks, logger, bar, mode)
    except FileNotFoundError as err:
        logger.critical(err.args[0])
        sys.exit(1)
    for (raw_path, unpacked_path) in failed_tasks:
        print(f'The creation of the root file failed for {raw_path}, '
              'removing {raw_path}')
        os.remove(raw_path)
        try:
            os.remove(unpacked_path)
        except FileNotFoundError:
            pass
    fracker_tasks = successful_tasks
    try:
        successful_fracker_tasks, failed_fracker_tasks = \
            frack_in_parallel(fracker_tasks, parallel_tasks,
                              logger, bar, mode, columns, compression)
    except FileNotFoundError as err:
        logger.critical(err.args[0])
        print('Unable to find compiled fracker, exiting ...')
        sys.exit(2)
    for (unpacked_path, formatted_path) in failed_fracker_tasks:
        print(f'The creation of the hdf file failed for {unpacked_path}, '
              'removing {unpacked_path}')
        os.remove(unpacked_path)
        try:
            os.remove(formatted_path)
        except FileNotFoundError:
            pass


_log_level_dict = {'DEBUG': logging.DEBUG,
                   'INFO': logging.INFO,
                   'WARNING': logging.WARNING,
                   'ERROR': logging.ERROR,
                   'CRITICAL': logging.CRITICAL}


@click.command
@click.argument('output_dir', type=click.Path(dir_okay=True),
                metavar='[Location containing the config and data]')
@click.option('--log', type=str, default=None,
              help='Enable logging and append logs to the filename passed to '
                   'this option')
@click.option('--loglevel', default='INFO',
              type=click.Choice(['DEBUG', 'INFO',
                                 'WARNING', 'ERROR', 'CRITICAL'],
                                case_sensitive=False))
@click.option('--root/--no-root', default=False,
              help='keep the rootfile generated as intermediary')
@click.option('--parallel_tasks', default=2, type=int,
              help='number of unpackers/frackers to run in parallel')
@click.option('--compression', '-c', default=3, type=int,
              help='Set the compression for the hdf file, 0 = no compression'
                   ' 9 = best compression')
def main(output_dir, log, loglevel, root, parallel_tasks, compression):
    if log is not None:
        logging.basicConfig(filename=log, level=_log_level_dict[loglevel],
                            format='[%(asctime)s] %(levelname)s:'
                                   '%(name)-50s %(message)s')
    output_dir = Path(output_dir)
    with open(output_dir / 'postprocessing_config.yaml') as pcf:
        procedure = yaml.safe_load(pcf.read())
        try:
            data_columns = procedure['data_columns']
        except KeyError:
            print('The procedure needs to specify data columns')
            sys.exit(1)
        try:
            mode = procedure['mode']
        except KeyError:
            mode = 'summary'
    default_config_file = output_dir / 'default_config.yaml'
    with open(default_config_file, 'r') as dfc:
        default_config = yaml.safe_load(dfc.read())
    initial_config_file = output_dir / 'initial_state_config.yaml'
    with open(initial_config_file, 'r') as icf:
        initial_config = yaml.safe_load(icf.read())
    full_config = dcu.update_dict(default_config, initial_config)
    run_config_paths = glob.glob(
            str(output_dir.absolute()) + '/run_*_config.yaml')
    run_raw_data_files = glob.glob(
            str(output_dir.absolute()) + '/run_*_data.raw')
    sorted(run_config_paths, key=lambda x: int(x.split('_')[-2]))
    sorted(run_raw_data_files, key=lambda x: int(x.split('_')[-2]))

    # prepare the data so that we can launch the conversions in parallel
    full_run_config_paths = []
    full_run_config_dir = os.path.dirname(run_config_paths[0])
    if procedure['diff']:
        bar = Bar('generating fully qualified configurations'.ljust(50, ' '),
                  max=len(run_config_paths))
        for run_config_file in run_config_paths:
            with open(run_config_file, 'r') as rcfgf:
                run_config_patch = yaml.safe_load(rcfgf.read())
            dcu.update_dict(full_config, run_config_patch, in_place=True)
            full_run_config_path = full_run_config_dir + '/' + \
                os.path.splitext(os.path.basename(run_config_file))[0] + \
                '_full.yaml'
            with open(full_run_config_path, 'w+') as frcf:
                frcf.write(yaml.dump(full_config))
            full_run_config_paths.append(Path(full_run_config_path))
            bar.next()
        bar.finish()
    else:
        full_run_config_paths = run_config_paths

    # actually do the postprocessing
    frack_data(run_raw_data_files,
               full_run_config_paths,
               mode,
               data_columns,
               parallel_tasks,
               bar=True,
               compression=compression
               )
    if not root:
        root_files = glob.glob(str(output_dir.absolute()) + '/*.root')
        for rf in root_files:
            logging.info(f'Deleting unneeded rootfile {rf}')
            os.remove(rf)
    for config_file in full_run_config_paths:
        logging.info(f'Deleting unneeded full_config_file {config_file}')
        os.remove(config_file)
    print('Done')
